"""Module for preservation of the legacy API, it will be removed in version 2.X of earthkit.transforms."""
from earthkit.transforms.climatology._aggregate import (
    anomaly,
    auto_anomaly,
    daily_max,
    daily_mean,
    daily_median,
    daily_min,
    daily_reduce,
    daily_std,
    max,
    mean,
    median,
    min,
    monthly_max,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_reduce,
    monthly_std,
    percentiles,
    quantiles,
    reduce,
    relative_anomaly,
    std,
)

__all__ = [
    reduce,
    mean,
    median,
    min,
    max,
    std,
    daily_reduce,
    daily_mean,
    daily_median,
    daily_min,
    daily_max,
    daily_std,
    monthly_reduce,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_max,
    monthly_std,
    quantiles,
    percentiles,
    anomaly,
    relative_anomaly,
    auto_anomaly,
]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Import and path setup ---------------------------------------------------

import datetime
import os
import sys

# This doesn't work, so need to find another way
# from earthkit import transforms
# version = transforms.__version__
# release = transforms.__version__

sys.path.insert(0, os.path.abspath("./"))
sys.path.insert(0, os.path.abspath("../"))

# -- Project information -----------------------------------------------------

project = "earthkit-transforms"
author = "European Centre for Medium Range Weather Forecasts"

year = datetime.datetime.now().year
years = "2022-%s" % (year,)
copyright = "%s, European Centre for Medium-Range Weather Forecasts (ECMWF)" % (years,)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "autoapi.extension",
    "myst_parser",
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]

# autodoc configuration
autodoc_typehints = "none"

# autoapi configuration
autoapi_dirs = ["../src/earthkit/transforms"]
autoapi_ignore = ["*/version.py", "earthkit/transforms/aggregate/"]
autoapi_options = [
    "members",
    "inherited-members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "imported-members",
]
autoapi_root = "_api"

# napoleon configuration
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_preprocess_types = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



# (C) Copyright 2024- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any, Callable, Optional, Union

import xarray as xr
from earthkit.transforms import _tools
from earthkit.transforms._aggregate import reduce as _reduce


@_tools.transform_inputs_decorator()
def reduce(
    dataarray: xr.DataArray | xr.Dataset,
    how: Union[str, Callable] = "mean",
    dim: Optional[str] = None,
):
    """Reduce data over the ensemble dimension.

    Parameters
    ----------
    dataarray : xr.DataArray | xr.Dataset
        The DataArray over which to calculate the climatological mean. Must
        contain a `time` dimension.
    how: str or callable
        Method used to reduce data. Default='mean', which will implement the xarray in-built mean.
        If string, it must be an in-built xarray reduce method, an earthkit how method or any numpy method.
        In the case of duplicate names, method selection is first in the order: xarray, earthkit, numpy.
        Otherwise it can be any function which can be called in the form `f(x, axis=axis, **kwargs)`
        to return the result of reducing an np.ndarray over an integer valued axis.
    dim : str (optional)
        Name of the ensemble dimension in the data object, default behaviour is to detect the
        ensemble dimension from the input object.
    """
    if dim is None:
        dim = _tools.get_dim_key(dataarray, "realization", raise_error=True)
    return _reduce(dataarray, how=how, dim=dim)


def mean(*args: Any, **kwargs: Any) -> xr.Dataset | xr.DataArray:
    """Calculate the ensemble mean.

    Parameters
    ----------
    dataarray : xr.DataArray | xr.Dataset
        The DataArray over which to calculate the climatological mean. Must
        contain a `time` dimension.
    dim : str (optional)
        Name of the ensemble dimension in the data object, default behaviour is to detect the
        ensemble dimension from the input object.
    *args, **kwargs
        Additional arguments and keyword arguments to pass to the underlying reduce function.
    """
    kwargs["how"] = "mean"
    return reduce(*args, **kwargs)


def std(*args: Any, **kwargs: Any) -> xr.Dataset | xr.DataArray:
    """Calculate the ensemble standard deviation.

    Parameters
    ----------
    dataarray : xr.DataArray | xr.Dataset
        The DataArray over which to calculate the climatological mean. Must
        contain a `time` dimension.
    dim : str (optional)
        Name of the ensemble dimension in the data object, default behaviour is to detect the
        ensemble dimension from the input object.
    *args, **kwargs
        Additional arguments and keyword arguments to pass to the underlying reduce function.
    """
    kwargs["how"] = "std"
    return reduce(*args, **kwargs)


def min(*args: Any, **kwargs: Any) -> xr.Dataset | xr.DataArray:
    """Calculate the ensemble minimum.

    Parameters
    ----------
    dataarray : xr.DataArray | xr.Dataset
        The DataArray over which to calculate the climatological mean. Must
        contain a `time` dimension.
    dim : str (optional)
        Name of the ensemble dimension in the data object, default behaviour is to detect the
        ensemble dimension from the input object.
    *args, **kwargs
        Additional arguments and keyword arguments to pass to the underlying reduce function.
    """
    kwargs["how"] = "min"
    return reduce(*args, **kwargs)


def max(*args: Any, **kwargs: Any) -> xr.Dataset | xr.DataArray:
    """Calculate the ensemble maximum.

    Parameters
    ----------
    dataarray : xr.DataArray | xr.Dataset
        The DataArray over which to calculate the climatological mean. Must
        contain a `time` dimension.
    dim : str (optional)
        Name of the ensemble dimension in the data object, default behaviour is to detect the
        ensemble dimension from the input object.
    *args, **kwargs
        Additional arguments and keyword arguments to pass to the underlying reduce function.
    """
    kwargs["how"] = "max"
    return reduce(*args, **kwargs)


def sum(*args: Any, **kwargs: Any) -> xr.Dataset | xr.DataArray:
    """Calculate the ensemble sum.

    Parameters
    ----------
    dataarray : xr.DataArray | xr.Dataset
        The DataArray over which to calculate the climatological sum. Must
        contain a `time` dimension.
    dim : str (optional)
        Name of the ensemble dimension in the data object, default behaviour is to detect the
        ensemble dimension from the input object.
    *args, **kwargs
        Additional arguments and keyword arguments to pass to the underlying reduce function.
    """
    kwargs["how"] = "sum"
    return reduce(*args, **kwargs)



import logging
import typing as T
from copy import deepcopy

import geopandas as gpd
import pandas as pd
import xarray as xr
from earthkit.transforms._tools import (
    array_namespace_from_object,
    ensure_list,
    get_how_xp,
    get_spatial_info,
    standard_weights,
    transform_inputs_decorator,
)
from numpy import ndarray

logger = logging.getLogger(__name__)


def _transform_from_latlon(lat, lon):
    """Return an Affine transformation of input 1D arrays of lat and lon.

    This assumes that both lat and lon are regular and contiguous.

    Parameters
    ----------
    lat : list
        arrays or lists of latitude and longitude

    lon : list
        arrays or lists of latitude and longitude
    """
    from affine import Affine

    trans = Affine.translation(lon[0] - (lon[1] - lon[0]) / 2, lat[0] - (lat[1] - lat[0]) / 2)
    scale = Affine.scale(lon[1] - lon[0], lat[1] - lat[0])

    return trans * scale


def rasterize(
    shape_list: T.List,
    coords: xr.core.coordinates.Coordinates,
    lat_key: str = "latitude",
    lon_key: str = "longitude",
    **kwargs,
) -> xr.DataArray:
    """Rasterize a list of geometries onto the given xarray coordinates.

    This only works for regular and contiguous latitude and longitude grids.

    Parameters
    ----------
    shape_list : affine.Affine
        List of geometries
    coords : xarray.coords
        Coordinates of dataarray to be masked
    lat_key :
        name of the latitude variables in the coordinates object
    lon_key :
        name of the longitude variables in the coordinates object
    dtype:
        datatype of the returned mask, default is `int`

    kwargs:
        Any other kwargs accepted by rasterio.features.rasterize

    Returns
    -------
    xr.DataArray
        A mask where points not inside the shape_list are set to `fill` value

    """
    from rasterio import features

    transform = _transform_from_latlon(coords[lat_key], coords[lon_key])
    out_shape = (len(coords[lat_key]), len(coords[lon_key]))
    raster = features.rasterize(shape_list, out_shape=out_shape, transform=transform, **kwargs)
    spatial_coords = {lat_key: coords[lat_key], lon_key: coords[lon_key]}
    return xr.DataArray(raster, coords=spatial_coords, dims=(lat_key, lon_key))


def mask_contains_points(
    shape_list: T.List,
    coords: xr.core.coordinates.Coordinates,
    lat_key: str = "latitude",
    lon_key: str = "longitude",
    **_kwargs,
) -> xr.DataArray:
    """Return a mask array for the spatial points of data that lie within shapes in shape_list.

    Function uses matplotlib.Path so can accept a list of points, this is much faster than shapely.
    It was initially included for use with irregular data but has been constructed to also accept
    regular data and return in the same format as the rasterize function.

    Parameters
    ----------
    shape_list :
        List of geometries
    coords : xarray.coords
        Coordinates of dataarray to be masked
    lat_key:
        name of the latitude variables in the coordinates object
    lon_key:
        name of the longitude variables in the coordinates object
    dtype:
        datatype of the returned mask, default is `int`

    Returns
    -------
    xr.DataArray
        A mask where points not inside the shape_list are set to `fill` value

    """
    import matplotlib.path as mpltPath

    xp = array_namespace_from_object(coords[lat_key])

    lat_dims = coords[lat_key].dims
    lon_dims = coords[lon_key].dims
    # Assert that latitude and longitude have the same dimensions
    #   (irregular data, e.g. x,y or obs)
    # or the dimensions are themselves (regular data) but we will probably
    # just use the rasterize function for the regular case
    assert (lat_dims == lon_dims) or (lat_dims == (lat_key,) and lon_dims == (lon_key,))
    if lat_dims == (lat_key,) and lon_dims == (lon_key,):
        lon_full, lat_full = xp.meshgrid(
            coords[lon_key].data,
            coords[lat_key].data,
        )
    else:
        lon_full, lat_full = (
            coords[lon_key].data,
            coords[lat_key].data,
        )
    # convert lat lon pairs to to points:
    points = list(
        zip(
            lon_full.flat,
            lat_full.flat,
        )
    )

    # get spatial dims and create output array:
    spatial_dims = list(set(lat_dims + lon_dims))
    outdata_shape = tuple(len(coords[dim]) for dim in spatial_dims)
    outdata = xp.full(outdata_shape, xp.nan)
    # loop over shapes and mask any point that is in the shape
    for shape in shape_list:
        for shp in shape[0]:
            shape_exterior = shp.exterior.coords.xy
            shape_exterior = list(
                zip(
                    list(shape_exterior[0]),  # longitudes
                    list(shape_exterior[1]),  # latitudes
                )
            )
            path = mpltPath.Path(shape_exterior)
            outdata.flat[path.contains_points(points)] = True

    out_coords = {coord: coords[coord] for coord in spatial_dims}
    outarray = xr.DataArray(outdata, coords=out_coords, dims=spatial_dims)

    return outarray


def _geopandas_to_shape_list(geodataframe):
    """Iterate over rows of a geodataframe."""
    return [row[1]["geometry"] for row in geodataframe.iterrows()]


def _array_mask_iterator(mask_arrays):
    """Iterate over mask arrays."""
    for mask_array in mask_arrays:
        yield mask_array > 0


def _shape_mask_iterator(shapes, target, regular=True, **kwargs):
    """Iterate over shape mask methods."""
    if isinstance(shapes, gpd.GeoDataFrame):
        shapes = _geopandas_to_shape_list(shapes)
    if regular:
        mask_function = rasterize
    else:
        mask_function = mask_contains_points
    for shape in shapes:
        shape_da = mask_function([shape], target.coords, **kwargs)
        yield shape_da


def shapes_to_masks(shapes: gpd.GeoDataFrame | list[gpd.GeoDataFrame], target, regular=True, **kwargs):
    """Create a list of masked dataarrays, if possible use the shape_mask_iterator.

    Parameters
    ----------
    shapes :
        A geodataframe or list of geodataframes containing the polygons for masks
    target :
        A dataarray to to create a mask for, only the geospatial coordinates are used

    regular :
        If True, data is on a regular grid so use rasterize method, if False use mask_contains_points
    all_touched :
        If True, all pixels touched by geometries will be considered in,
        if False, only pixels whose center is within. Default is False.
        Only valid for regular data.

    kwargs:
        kwargs accepted by the masking methods, rasterize or mask_contains_points

    Returns
    -------
    list[xr.DataArray]
        A list of masks where points inside each geometry are 1, and those outside are xp.nan

    """
    if isinstance(shapes, gpd.GeoDataFrame):
        shapes = _geopandas_to_shape_list(shapes)
    if regular:
        mask_function = rasterize
    else:
        mask_function = mask_contains_points

    return [mask_function([shape], target.coords, **kwargs) for shape in shapes]


def shapes_to_mask(shapes, target, regular=True, **kwargs):
    """Create a single masked dataarray based on all features in shapes.

    If possible use the shape_mask_iterator.

    Parameters
    ----------
    shapes :
        A geodataframe or list of geodataframes containing the polygons for masks
    target :
        A dataarray to to create a mask for, only the geospatial coordinates are used

    regular :
        If True, data is on a regular grid so use rasterize method,
        if False use mask_contains_points
    all_touched :
        If True, all pixels touched by geometries will be considered in,
        if False, only pixels whose center is within. Default is False.
        Only valid for regular data.

    kwargs:
        kwargs accepted by the masking methods, rasterize or mask_contains_points

    Returns
    -------
    xr.DataArray
        A mask where points inside any geometry are 1, and those outside are xp.nan

    """
    if isinstance(shapes, gpd.GeoDataFrame):
        shapes = _geopandas_to_shape_list(shapes)
    if regular:
        mask_function = rasterize
    else:
        mask_function = mask_contains_points

    return mask_function(shapes, target.coords, **kwargs)


def get_mask_dim_index(
    mask_dim: str | None | T.Dict[str, T.Any],
    geodataframe: gpd.geodataframe.GeoDataFrame,
    default_index_name: str = "index",
):
    if isinstance(mask_dim, str):
        if mask_dim in geodataframe:
            mask_dim_index = pd.Index(geodataframe[mask_dim])
        else:
            mask_dim_index = geodataframe.index.rename(mask_dim)
    elif isinstance(mask_dim, dict):
        assert len(mask_dim) == 1, "If provided as a dictionary, mask_dim should have only one key value pair"
        _mask_dim, _mask_dim_values = mask_dim.items()
        mask_dim_index = pd.Index(_mask_dim_values, name=_mask_dim)
    elif mask_dim is None:
        # Use the index of the data frame
        mask_dim_index = geodataframe.index
        if mask_dim_index.name is None:
            mask_dim_index = mask_dim_index.rename(default_index_name)
    else:
        raise ValueError("Unrecognised mask_dim format")

    return mask_dim_index


@transform_inputs_decorator()
def mask(
    dataarray: xr.Dataset | xr.DataArray,
    geodataframe: gpd.geodataframe.GeoDataFrame,
    mask_dim: str | None = None,
    lat_key: str | None = None,
    lon_key: str | None = None,
    chunk: bool = True,
    union_geometries: bool = False,
    **mask_kwargs,
) -> xr.Dataset | xr.DataArray:
    """Apply multiple shape masks to some gridded data.

    Each feature in shape is treated as an individual mask to apply to
    data. The data provided is returned with an additional dimension equal in
    length to the number of features in the shape object, this can result in very
    large files which will slow down your script. It may be better to loop
    over individual features, or directly apply the mask with the shapes.reduce.

    Parameters
    ----------
    dataarray :
        Xarray data object (must have geospatial coordinates).
    geodataframe :
        Geopandas Dataframe containing the polygons for aggregations
    mask_dim :
        dimension that will be created to accomodate the masked arrays, default is the index
        of the geodataframe
    all_touched :
        If True, all pixels touched by geometries will be considered in,
        if False, only pixels whose center is within. Default is False.
        Only valid for regular data.
    lat_key :
        key for latitude variable, default behaviour is to detect variable keys.
    lon_key :
        key for longitude variable, default behaviour is to detect variable keys.
    chunk :  bool
        Boolean to indicate whether to use chunking, default = `True`.
        This is advised as spatial.masks can create large results. If you are working with small
        arrays, or you have implemented you own chunking rules you may wish to disable it.
    union_geometries : bool
        Boolean to indicate whether to union all geometries before masking.
        Default is `False`, which will apply each geometry in the geodataframe as a separate mask.
    mask_kwargs :
        Any kwargs to pass into the mask method

    Returns
    -------
    xr.Dataset | xr.DataArray
        A masked data array with dimensions [feautre_id] + [data.dims].
        Each slice of layer corresponds to a feature in layer.

    """
    spatial_info = get_spatial_info(dataarray, lat_key=lat_key, lon_key=lon_key)
    # Get spatial info required by mask functions:
    mask_kwargs = {**mask_kwargs, **{key: spatial_info[key] for key in ["lat_key", "lon_key", "regular"]}}
    mask_dim_index = get_mask_dim_index(mask_dim, geodataframe)

    if union_geometries:
        loop_masks = [shapes_to_mask(geodataframe, dataarray, **mask_kwargs)]
    else:
        loop_masks = _shape_mask_iterator(geodataframe, dataarray, **mask_kwargs)

    masked_arrays = []
    for this_mask in loop_masks:
        this_masked_array = dataarray.where(this_mask)
        if chunk:
            this_masked_array = this_masked_array.chunk()
        masked_arrays.append(this_masked_array.copy())

    if union_geometries:
        out = masked_arrays[0]
    else:
        # TODO: remove ignore type if xarray concat typing is updated
        out = xr.concat(masked_arrays, dim=mask_dim_index.name)  # type: ignore
        if chunk:
            out = out.chunk({mask_dim_index.name: 1})

        out = out.assign_coords({mask_dim_index.name: mask_dim_index})

    out.attrs.update(geodataframe.attrs)

    return out


@transform_inputs_decorator()
def reduce(
    dataarray: xr.Dataset | xr.DataArray,
    geodataframe: gpd.GeoDataFrame | None = None,
    mask_arrays: xr.DataArray | list[xr.DataArray] | None = None,
    **kwargs,
) -> xr.Dataset | xr.DataArray:
    """Apply a shape object to an xarray.DataArray object using the specified 'how' method.

    Geospatial coordinates are reduced to a dimension representing the list of features in the shape object.

    Parameters
    ----------
    dataarray :
        Xarray data object (must have geospatial coordinates).
    geodataframe :
        Geopandas Dataframe containing the polygons for aggregations
    mask_arrays :
        precomputed mask array[s], if provided this will be used instead of creating a new mask.
        They must be on the same spatial grid as the dataarray.
    how :
        method used to apply mask. Default='mean', which calls xp.nanmean
    weights :
        Provide weights for aggregation, also accepts recognised keys for weights, e.g.
        'latitude'
    lat_key/lon_key :
        key for latitude/longitude variable, default behaviour is to detect variable keys.
    extra_reduce_dims :
        any additional dimensions to aggregate over when reducing over spatial dimensions
    mask_dim :
        dimension that will be created after the reduction of the spatial dimensions, default is the index
        of the dataframe
    all_touched :
        If True, all pixels touched by geometries will be considered in,
        if False, only pixels whose center is within. Default is False.
        Only valid for regular data.
    mask_kwargs :
        Any kwargs to pass into the mask method
    mask_arrays :
        precomputed mask array[s], if provided this will be used instead of creating a new mask.
        They must be on the same spatial grid as the dataarray.
    return_as :
        what format to return the data object, `pandas` or `xarray`. Work In Progress
    compact :
        If True, return a compact pandas.DataFrame with the reduced data as a new column.
        If False, return a fully expanded pandas.DataFrame.
        Only valid if return_as is `pandas`
    how_label :
        label to append to variable name in returned object, default is not to append
    kwargs :
        kwargs recognised by the how function

    Returns
    -------
    xr.Dataset | xr.DataArray
        A data array with dimensions `features` + `data.dims not in 'lat','lon'`.
        Each slice of layer corresponds to a feature in layer.

    """
    assert not (
        geodataframe is not None and mask_arrays is not None
    ), "Either a geodataframe or mask arrays must be provided, not both"
    if mask_arrays is not None:
        _mask_arrays: list[xr.DataArray] | None = ensure_list(mask_arrays)
    else:
        _mask_arrays = None

    if isinstance(dataarray, xr.Dataset):
        return_as: str = kwargs.pop("return_as", "xarray")
        if return_as in ["xarray"]:
            out_ds = xr.Dataset().assign_attrs(dataarray.attrs)
            for var in dataarray.data_vars:
                out_da = _reduce_dataarray_as_xarray(
                    dataarray[var], geodataframe=geodataframe, mask_arrays=_mask_arrays, **kwargs
                )
                out_ds[out_da.name] = out_da
            return out_ds
        elif "pandas" in return_as:
            logger.warning(
                "Returning reduced data in pandas format is considered "
                "experimental and may change in future"
                "versions of earthkit"
            )
            if geodataframe is not None:
                out = geodataframe
                for var in dataarray.data_vars:
                    out = _reduce_dataarray_as_pandas(dataarray[var], geodataframe=out, **kwargs)
            else:
                out = None
                for var in dataarray.data_vars:
                    _out = _reduce_dataarray_as_pandas(dataarray[var], mask_arrays=_mask_arrays, **kwargs)
                    if out is None:
                        out = _out
                    else:
                        out = pd.merge(out, _out)
            return out
        else:
            raise TypeError("Return as type not recognised or incompatible with inputs")
    else:
        return _reduce_dataarray_as_xarray(
            dataarray, geodataframe=geodataframe, mask_arrays=_mask_arrays, **kwargs
        )


def _reduce_dataarray_as_xarray(
    dataarray: xr.DataArray,
    geodataframe: gpd.GeoDataFrame | None = None,
    mask_arrays: list[xr.DataArray] | None = None,
    how: T.Callable | str = "mean",
    weights: None | str | ndarray = None,
    lat_key: str | None = None,
    lon_key: str | None = None,
    extra_reduce_dims: list | str = [],
    mask_dim: str | None = None,
    how_label: str | None = None,
    squeeze: bool = True,
    all_touched: bool = False,
    mask_kwargs: dict[str, T.Any] = dict(),
    return_geometry_as_coord: bool = False,
    **reduce_kwargs,
) -> xr.DataArray:
    """Reduce an xarray.DataArray object over its geospatial dimensions using the specified 'how' method.

    If a geodataframe is provided the DataArray is reduced over each feature in the geodataframe.
    Geospatial coordinates are reduced to a dimension representing the list of features in the shape object.

    Parameters
    ----------
    dataarray :
        Xarray data object (must have geospatial coordinates).
    geodataframe :
        Geopandas Dataframe containing the polygons for aggregations
    mask_arrays :
        precomputed mask array[s], if provided this will be used instead of creating a new mask.
        They must be on the same spatial grid as the dataarray.
    how :
        method used to apply mask. Default='mean', which calls xp.nanmean
    weights :
        Provide weights for aggregation, also accepts recognised keys for weights, e.g.
        'latitude'
    lat_key :
        key for latitude variable, default behaviour is to detect variable keys.
    lon_key :
        key for longitude variable, default behaviour is to detect variable keys.
    extra_reduce_dims :
        any additional dimensions to aggregate over when reducing over spatial dimensions
    mask_dim :
        dimension that will be created after the reduction of the spatial dimensions, default = `"index"`
    return_as :
        what format to return the data object, `"pandas"` or `"xarray"`. Work In Progress
    how_label :
        label to append to variable name in returned object, default is `how`
    mask_kwargs :
        Any kwargs to pass into the mask method
    reduce_kwargs :
        kwargs recognised by the how function
    return_geometry_as_coord :
        include the geometries as a coordinate in the returned xarray object. WARNING: geometries are not
        serialisable objects, therefore this xarray will not be saveable as netCDF.
    all_touched :
        If True, all pixels touched by geometries will be considered in,
        if False, only pixels whose center is within. Default is False.
        Only valid for regular data.
    squeeze :
        If True, squeeze the output xarray object, default is True

    Returns
    -------
    xr.DataArray
        A data array with dimensions [features] + [data.dims not in ['lat','lon']].
        Each slice of layer corresponds to a feature in layer

    """
    xp = array_namespace_from_object(dataarray)
    extra_out_attrs = {}
    how_str: None | str = None
    if weights is None:
        # convert how string to a method to apply
        if isinstance(how, str):
            how_str = deepcopy(how)
            how = reduce_how = get_how_xp(how, xp=xp)
        else:
            reduce_how = how
        assert callable(how), f"how must be a callable: {how}"
        if how_str is None:
            # get label from how method
            how_str = how.__name__
    else:
        # Create any standard weights, e.g. latitude.
        # TODO: handle kwargs better, currently only lat_key is accepted
        if isinstance(weights, str):
            _weights = standard_weights(dataarray, weights, lat_key=lat_key)
        else:
            _weights = weights
        # We ensure the callable is a string
        if callable(how):
            how = weighted_how = how.__name__
        else:
            weighted_how = how
        if how_str is None:
            how_str = how

    how_str = how_str or how_label
    new_long_name_components = [
        comp for comp in [how_str, dataarray.attrs.get("long_name", dataarray.name)] if comp is not None
    ]
    new_long_name = " ".join(new_long_name_components)
    extra_out_attrs.update({"long_name": new_long_name})
    new_short_name_components = [f"{comp}" for comp in [dataarray.name, how_label] if comp is not None]
    new_short_name = "_".join(new_short_name_components)

    if isinstance(extra_reduce_dims, str):
        extra_reduce_dims = [extra_reduce_dims]

    spatial_info = get_spatial_info(dataarray, lat_key=lat_key, lon_key=lon_key)
    # Get spatial info required by mask functions:
    mask_kwargs = {**mask_kwargs, **{key: spatial_info[key] for key in ["lat_key", "lon_key", "regular"]}}
    # All touched only valid for rasterize method
    if spatial_info["regular"]:
        mask_kwargs.setdefault("all_touched", all_touched)
    else:
        if all_touched:
            logger.warning("all_touched only valid for regular data, ignoring")
    spatial_dims = spatial_info.get("spatial_dims")

    reduce_dims = spatial_dims + extra_reduce_dims
    extra_out_attrs.update({"reduce_dims": reduce_dims})
    reduce_kwargs.update({"dim": reduce_dims})
    # If using a pre-computed mask arrays,
    # then iterator is just dataarray*mask_array
    if mask_arrays is not None:
        masked_data_list = _array_mask_iterator(mask_arrays)
    else:
        # If no geodataframe, then no mask, so create a dummy mask:
        if geodataframe is None:
            masked_data_list = [xr.ones_like(dataarray)]
        else:
            masked_data_list = _shape_mask_iterator(geodataframe, dataarray, **mask_kwargs)

    reduced_list = []
    for masked_data in masked_data_list:
        this = dataarray.where(masked_data, other=xp.nan)

        # If weighted, use xarray weighted arrays which
        # correctly handle missing values etc.
        if weights is not None:
            this_weighted = this.weighted(_weights)
            reduced_list.append(this_weighted.__getattribute__(weighted_how)(**reduce_kwargs))
        else:
            reduced = this.reduce(reduce_how, **reduce_kwargs).compute()
            reduced = reduced.assign_attrs(dataarray.attrs)
            reduced_list.append(reduced)

    if squeeze:
        reduced_list = [red_data.squeeze() for red_data in reduced_list]

    # If no geodataframe, there is just one reduced array
    if geodataframe is not None:
        mask_dim_index = get_mask_dim_index(mask_dim, geodataframe)
        out_xr = xr.concat(reduced_list, dim=mask_dim_index)
    elif mask_dim is None and len(reduced_list) == 1:
        out_xr = reduced_list[0]
    else:
        _concat_dim_name = mask_dim or "index"
        out_xr = xr.concat(reduced_list, dim=_concat_dim_name)

    out_xr = out_xr.rename(new_short_name)
    if geodataframe is not None:
        if return_geometry_as_coord:
            out_xr = out_xr.assign_coords(
                **{"geometry": (mask_dim_index.name, [_g for _g in geodataframe["geometry"]])}
            )
        out_xr = out_xr.assign_attrs({**geodataframe.attrs, **extra_out_attrs})

    return out_xr


def _reduce_dataarray_as_pandas(
    dataarray: xr.DataArray, geodataframe: gpd.GeoDataFrame | None = None, compact: bool = False, **kwargs
) -> pd.DataFrame:
    """Reduce an xarray.DataArray object over its geospatial dimensions using the specified 'how' method.

    If a geodataframe is provided the DataArray is reduced over each feature in the geodataframe.
    Geospatial coordinates are reduced to a dimension representing the list of features in the shape object.

    Parameters
    ----------
    dataarray :
        Xarray data object (must have geospatial coordinates).
    geodataframe :
        Geopandas Dataframe containing the polygons for aggregations
    compact :
        If True, return a compact pandas.DataFrame with the reduced data as a new column.
        If False, return a fully expanded pandas.DataFrame.
    kwargs :
        kwargs accepted by the :function:_reduce_dataarray_as_xarray function

    Returns
    -------
    pd.DataFrame
        A pandas.DataFrame similar to the geopandas dataframe, with the reduced data
        added as a new column.

    """
    out_xr = _reduce_dataarray_as_xarray(dataarray, geodataframe=geodataframe, **kwargs)

    reduce_attrs = {f"{dataarray.name}": dataarray.attrs, f"{out_xr.name}": out_xr.attrs}

    if geodataframe is None:
        mask_dim = kwargs.get("mask_dim", "index")
        if mask_dim not in out_xr.dims:
            out_xr = xr.concat([out_xr], dim=mask_dim)
        # If no geodataframe, then just convert xarray to dataframe
        out = out_xr.to_dataframe()
        # Add attributes to the dataframe
        out.attrs.update({"reduce_attrs": reduce_attrs})
        return out

    # Otherwise, splice the geodataframe and reduced xarray
    reduce_attrs = {
        **geodataframe.attrs.get("reduce_attrs", {}),
        **reduce_attrs,
    }

    # TODO: somehow remove repeat call of get_mask_dim_index (see _reduce_dataarray_as_xarray)
    mask_dim_index = get_mask_dim_index(kwargs.get("mask_dim"), geodataframe)
    mask_dim_name = mask_dim_index.name
    out = geodataframe.set_index(mask_dim_index)
    if mask_dim_name not in out_xr.dims:
        out_xr = xr.concat([out_xr], dim=mask_dim_name)
    if not compact:  # Return as a fully expanded pandas.DataFrame
        # Convert to DataFrame
        out = out.join(out_xr.to_dataframe())
    else:
        # add the reduced data into a new column as a numpy array,
        # store the dim information in the attributes
        _out_dims = [str(dim) for dim in dataarray.coords if dim in out_xr.dims]
        out_dims = {dim: dataarray[dim].data for dim in _out_dims}
        reduce_attrs[f"{out_xr.name}"].update({"dims": out_dims})
        reduced_list = [
            out_xr.sel(**{mask_dim_name: mask_dim_value}).data
            for mask_dim_value in out_xr[mask_dim_name].data
        ]
        out = out.assign(**{f"{out_xr.name}": reduced_list})

    out.attrs.update({"reduce_attrs": reduce_attrs})

    return out



from typing import Any

import numpy as np

try:
    import cupy as cp
except ImportError:
    cp = None
import pandas as pd
import pytest
import xarray as xr
from earthkit.transforms._tools import (
    get_dim_key,
    get_how,
    get_how_xp,
    get_spatial_info,
    groupby_kwargs_decorator,
    latitude_weights,
    nanaverage,
    time_dim_decorator,
)


# Define a dummy function to decorate
def dummy_func(dataarray, *args, time_dim=None, **kwargs):
    return dataarray, time_dim


# Define a dummy function to decorate
def dummy_func2(dataarray, *args, time_dim=None, **kwargs):
    return dataarray


# Test case for the decorator when time_shift is provided
def test_time_dim_decorator_time_shift_provided():
    # Prepare test data
    dataarray = xr.DataArray(
        [1, 2, 3], dims=["time"], coords={"time": pd.date_range("2000-01-01", periods=3)}
    )

    # Call the decorated function with time_shift provided
    result, result_time_dim = time_dim_decorator(dummy_func)(dataarray, time_shift={"days": 1})

    assert result_time_dim == "time"

    # Check if the time dimension is correctly shifted
    expected_coords = pd.date_range("2000-01-02", periods=3)
    assert all(result.coords["time"].values == expected_coords)


# Test case for the decorator when both time_dim and time_shift are provided
def test_time_dim_decorator_time_dim_and_time_shift_provided():
    # Prepare test data
    dataarray = xr.DataArray(
        [1, 2, 3], dims=["dummy"], coords={"dummy": pd.date_range("2000-01-01", periods=3)}
    )

    # Call the decorated function with both time_dim and time_shift provided
    result, result_time_dim = time_dim_decorator(dummy_func)(
        dataarray, time_dim="dummy", time_shift={"days": 1}
    )

    # Check if the time dimension remains unchanged
    assert result_time_dim == "dummy"

    # Check if the time dimension is correctly shifted
    expected_coords = pd.date_range("2000-01-02", periods=3)
    assert all(result.coords["dummy"].values == expected_coords)


# Test case for the decorator when both time_dim and time_shift are provided
def test_time_dim_decorator_not_found_error():
    dataarray = xr.DataArray([1, 2, 3], dims=["dummy"], coords={"dummy": [1, 2, 3]})

    # Check error raised when cannot find time dimension
    with pytest.raises(KeyError):
        time_dim_decorator(dummy_func)(dataarray, time_shift={"days": 1})


# Test case for the decorator when time_shift is provided and remove_partial_periods=True
def test_time_dim_decorator_time_shift_provided_trim_shifted():
    # Prepare test data
    dataarray = xr.DataArray(
        [1, 2, 3], dims=["time"], coords={"time": pd.date_range("2000-01-01", periods=3)}
    )

    # Call the decorated function with time_shift provided
    result = time_dim_decorator(dummy_func2)(dataarray, time_shift={"days": 1}, remove_partial_periods=True)
    # Check if the time dimension is correctly shifted, and the start and end dates are removed
    expected_coords = pd.date_range("2000-01-03", periods=1)
    assert all(result.coords["time"].values == expected_coords)


# Define a dummy function to decorate
def gb_dummy_func(*args, groupby_kwargs=None, **kwargs):
    return groupby_kwargs, kwargs


# Test case for the decorator when groupby_kwargs is None
def test_groupby_kwargs_decorator_none():
    # Call the decorated function with no groupby_kwargs
    result_groupby_kwargs, result_kwargs = groupby_kwargs_decorator(gb_dummy_func)()

    # Check if groupby_kwargs is None and other kwargs are empty
    assert result_groupby_kwargs == {}
    assert result_kwargs == {}


# Test case for the decorator when groupby_kwargs is provided
def test_groupby_kwargs_decorator_provided():
    # Prepare groupby_kwargs and other kwargs
    groupby_kwargs = {"frequency": "day", "bin_widths": 1}
    other_kwargs = {"method": "linear", "fill_value": 0}

    # Call the decorated function with groupby_kwargs provided
    result_groupby_kwargs, result_kwargs = groupby_kwargs_decorator(gb_dummy_func)(
        **groupby_kwargs, **other_kwargs
    )

    assert result_groupby_kwargs == groupby_kwargs
    assert result_kwargs == other_kwargs


# Test case for the decorator when some groupby_kwargs are provided as keyword arguments
def test_groupby_kwargs_decorator_partial_provided():
    # Prepare groupby_kwargs and other kwargs
    groupby_kwargs = {"frequency": "day", "bin_widths": 1}
    other_kwargs = {"method": "linear"}

    # Call the decorated function with some groupby_kwargs provided as keyword arguments
    result_groupby_kwargs, result_kwargs = groupby_kwargs_decorator(gb_dummy_func)(
        **groupby_kwargs, **other_kwargs
    )

    assert result_groupby_kwargs == groupby_kwargs
    assert result_kwargs == other_kwargs


# Test case for the decorator when groupby_kwargs is provided
def test_groupby_kwargs_decorator_override():
    # Prepare groupby_kwargs and other kwargs
    groupby_kwargs = {"frequency": "day", "bin_widths": 1}
    other_kwargs = {"method": "linear", "fill_value": 0}

    override_groupby_kwargs = {"frequency": "hour"}

    # Call the decorated function with groupby_kwargs provided
    result_groupby_kwargs, result_kwargs = groupby_kwargs_decorator(gb_dummy_func)(
        groupby_kwargs=override_groupby_kwargs, **groupby_kwargs, **other_kwargs
    )

    groupby_kwargs.update(override_groupby_kwargs)
    assert result_groupby_kwargs == groupby_kwargs
    assert result_kwargs == other_kwargs


@pytest.mark.parametrize(
    "how_str, how_expected",
    (
        ["mean", np.nanmean],
        ["numpy.mean", np.mean],
        ["numpy.nanmean", np.nanmean],
        ["nanaverage", nanaverage],
        ["average", nanaverage],
        ["numpy.average", np.average],
        ["stddev", np.nanstd],
        ["std", np.nanstd],
        ["stdev", np.nanstd],
        ["sum", np.nansum],
        ["max", np.nanmax],
        ["min", np.nanmin],
        ["median", np.nanmedian],
        ["q", np.nanquantile],
        ["quantile", np.nanquantile],
        ["percentile", np.nanpercentile],
        ["p", np.nanpercentile],
    ),
)
def test_get_how(how_str: str, how_expected: Any):
    assert how_expected == get_how(how_str)


def test_get_how_xp_np():
    test_case = (
        ["mean", np.nanmean],
        ["numpy.mean", np.mean],
        ["numpy.nanmean", np.nanmean],
        ["nanaverage", nanaverage],
        ["average", nanaverage],
        ["numpy.average", np.average],
        ["stddev", np.nanstd],
        ["std", np.nanstd],
        ["stdev", np.nanstd],
        ["sum", np.nansum],
        ["max", np.nanmax],
        ["min", np.nanmin],
        ["median", np.nanmedian],
        ["q", np.nanquantile],
        ["quantile", np.nanquantile],
        ["percentile", np.nanpercentile],
        ["p", np.nanpercentile],
    )
    for how_str, expected in test_case:
        assert expected == get_how_xp(how_str, xp=np)


@pytest.mark.parametrize(
    "data_object",
    (
        xr.DataArray(np.random.rand(10, 10), dims=["x", "y"]),
        xr.Dataset(
            {"var": (["x", "y"], np.random.rand(10, 10))}, coords={"x": np.arange(10), "y": np.arange(10)}
        ),
        np.random.rand(10, 10),
    ),
)
def test_get_how_xp_np_objects(data_object):
    test_cases = (
        ["mean", np.nanmean],
        ["numpy.mean", np.mean],
        ["numpy.nanmean", np.nanmean],
        ["nanaverage", nanaverage],
        ["average", nanaverage],
        ["numpy.average", np.average],
        ["stddev", np.nanstd],
        ["std", np.nanstd],
        ["stdev", np.nanstd],
        ["sum", np.nansum],
        ["max", np.nanmax],
        ["min", np.nanmin],
        ["median", np.nanmedian],
        ["q", np.nanquantile],
        ["quantile", np.nanquantile],
        ["percentile", np.nanpercentile],
        ["p", np.nanpercentile],
    )
    for how_str, expected in test_cases:
        assert expected == get_how_xp(how_str, data_object=data_object)


@pytest.mark.skipif(cp is None, reason="Cupy is not installed")
def test_get_how_xp_cp():
    test_cases = [
        ("mean", cp.nanmean),
        ("numpy.mean", cp.mean),
        ("numpy.nanmean", cp.nanmean),
        ("nanaverage", nanaverage),
        ("average", nanaverage),
        ("numpy.average", cp.average),
        ("stddev", cp.nanstd),
        ("std", cp.nanstd),
        ("stdev", cp.nanstd),
        ("sum", cp.nansum),
        ("max", cp.nanmax),
        ("min", cp.nanmin),
        ("median", cp.nanmedian),
        ("q", cp.nanquantile),
        ("quantile", cp.nanquantile),
        ("percentile", cp.nanpercentile),
        ("p", cp.nanpercentile),
    ]
    for how_str, expected in test_cases:
        assert get_how_xp(how_str, xp=cp) == expected


# Define a helper function to create dummy dataarray with given axis attribute
def create_dataarray_with_axis(axes, dims):
    dim_arrs = {}
    for axis, dim in zip(axes, dims):
        dim_arrs[dim] = xr.DataArray([1, 2], dims=dim)
        if axis is not None:
            dim_arrs[dim] = dim_arrs[dim].assign_attrs({"axis": axis})

    return xr.DataArray(np.ones([2] * len(dims)), dims=dims, coords=dim_arrs)


@pytest.mark.parametrize(
    "axis, dim_expected",
    (
        ["x", "lon"],
        ["y", "lat"],
        ["t", "time"],
        ["z", "z"],
        ["realization", "realization"],
    ),
)
# Test case for the function when axis matches an attribute in the dataset's dimensions
def test_get_dim_keys(axis, dim_expected):
    # Create a dataarray with an axis attribute
    dataarray = create_dataarray_with_axis(["x", "y", "t", None], ["lon", "lat", "time", "realization"])
    # Check if the correct dimension key is returned
    assert dim_expected == get_dim_key(dataarray, axis)
    # Create a dataarray with an axis attribute
    dataarray = create_dataarray_with_axis([None, None, None, None], ["lon", "lat", "time", "realization"])
    # Check if the correct dimension key is returned
    assert dim_expected == get_dim_key(dataarray, axis)


# Test case for the function when axis matches an attribute in the dataset's dimensions
def test_get_dim_keys_raises():
    axis = "z"
    # Create a dataarray with an axis attribute
    dataarray = create_dataarray_with_axis(["x", None], ["lon", "realization"])
    # Check if the correct dimension key is returned
    with pytest.raises(ValueError, match=f"Unable to find dimension key for axis '{axis}' in dataarray"):
        get_dim_key(dataarray, axis, raise_error=True)


# Test case for the function when axis matches an attribute in the dataset's dimensions
def test_get_spatial_info():
    # Create a dataarray with an axis attribute
    dataarray = create_dataarray_with_axis(["x", "y", "t"], ["lon", "lat", "time"])
    # Check if the correct dimension key is returned
    expected_result = {"lat_key": "lat", "lon_key": "lon", "regular": True, "spatial_dims": ["lat", "lon"]}
    assert expected_result == get_spatial_info(dataarray)


def test_latitude_weights():
    da = xr.DataArray(np.arange(5), dims=("y"), coords={"y": [-90, -60, 0, 60, 90]})

    weights = latitude_weights(da)
    assert np.allclose(weights, [0, 0.5, 1, 0.5, 0])

    weights = latitude_weights(da.rename({"y": "latitude"}))
    assert np.allclose(weights, [0, 0.5, 1, 0.5, 0])



import pytest

# from earthkit.data.core.temporary import temp_directory
import xarray as xr
from earthkit import data as ek_data
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.transforms import climatology

# Use caching for speedy repeats
ek_data.settings.set("cache-policy", "user")


def get_data():
    remote_era5_file = earthkit_remote_test_data_file("era5_temperature_france_2015_2016_2017_3deg.grib")
    return ek_data.from_source("url", remote_era5_file)


@pytest.mark.parametrize(
    "method, how",
    (
        (climatology.monthly_mean, "mean"),
        (climatology.monthly_median, "median"),
        (climatology.monthly_min, "min"),
        (climatology.monthly_max, "max"),
    ),
)
@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_climatology_monthly(in_data, expected_return_type, method, how):
    clim_m = method(in_data)
    assert isinstance(clim_m, expected_return_type)
    assert "month" in list(clim_m.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == clim_m.name
    else:
        assert "2t" in clim_m

    clim_m = method(in_data, how_label=how)
    assert isinstance(clim_m, expected_return_type)
    assert "month" in list(clim_m.dims)
    if expected_return_type == xr.DataArray:
        assert f"2t_{how}" == clim_m.name
    else:
        assert f"2t_{how}" in clim_m


@pytest.mark.parametrize(
    "method, how",
    (
        (climatology.daily_mean, "mean"),
        # (climatology.daily_median, "median"),
        # (climatology.daily_min, "min"),
        (climatology.daily_max, "max"),
    ),
)
@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        # [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_climatology_daily(in_data, expected_return_type, method, how):
    clim_d = method(in_data)
    assert isinstance(clim_d, expected_return_type)
    assert "dayofyear" in list(clim_d.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == clim_d.name
    else:
        assert "2t" in clim_d

    clim_d = method(in_data, how_label=how)
    assert isinstance(clim_d, expected_return_type)
    assert "dayofyear" in list(clim_d.dims)
    if expected_return_type == xr.DataArray:
        assert f"2t_{how}" == clim_d.name
    else:
        assert f"2t_{how}" in clim_d


@pytest.mark.parametrize(
    "clim_method",
    (
        climatology.monthly_mean,
        climatology.monthly_median,
    ),
)
@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        # [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_anomaly_monthly(in_data, expected_return_type, clim_method):
    clim_m = clim_method(in_data)
    anom_m = climatology.anomaly(in_data, clim_m, frequency="month")
    assert isinstance(anom_m, expected_return_type)
    if expected_return_type == xr.DataArray:
        assert "2t" == anom_m.name
    else:
        assert "2t" in anom_m
    anom_m = climatology.anomaly(in_data, clim_m, frequency="month", how_label="anomaly")
    assert isinstance(anom_m, expected_return_type)
    if expected_return_type == xr.DataArray:
        assert "2t_anomaly" == anom_m.name
    else:
        assert "2t_anomaly" in anom_m


@pytest.mark.parametrize(
    "clim_method",
    (
        climatology.daily_mean,
        climatology.daily_median,
    ),
)
@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        # [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_anomaly_daily(in_data, expected_return_type, clim_method):
    clim_d = clim_method(in_data)
    anom_d = climatology.anomaly(in_data, clim_d, frequency="dayofyear")
    assert isinstance(anom_d, expected_return_type)
    if expected_return_type == xr.DataArray:
        assert "2t" == anom_d.name
    else:
        assert "2t" in anom_d
    anom_d = climatology.anomaly(in_data, clim_d, frequency="dayofyear", how_label="anomaly")
    assert isinstance(anom_d, expected_return_type)
    if expected_return_type == xr.DataArray:
        assert "2t_anomaly" == anom_d.name
    else:
        assert "2t_anomaly" in anom_d



import typing as T

import xarray as xr
from earthkit.transforms import _tools

# Use numpy.ndarray for Array type hinting
from numpy import ndarray as Array


def how_label_rename(
    dataarray: xr.Dataset | xr.DataArray,
    how_label: str | None = None,
) -> xr.Dataset | xr.DataArray:
    if how_label is not None:
        # Update variable names, depends on dataset or dataarray format
        if isinstance(dataarray, xr.Dataset):
            renames = {data_arr: f"{data_arr}_{how_label}" for data_arr in dataarray}
            dataarray = dataarray.rename(renames)
        else:
            dataarray = dataarray.rename(f"{dataarray.name}_{how_label}")

    return dataarray


def _reduce_dataarray(
    dataarray: xr.DataArray,
    how: T.Callable | str = "mean",
    weights: None | str | Array = None,
    how_label: str | None = None,
    how_dropna: bool = False,
    xp: T.Any = None,
    **kwargs,
):
    """Reduce an xarray.dataarray or xarray.dataset using a specified `how` method.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray
        Data object to reduce
    how: str or callable
        Method used to reduce data. Default='mean', which will implement the xarray in-built mean.
        If string, it must be an in-built xarray reduce method, a earthkit how method or any numpy method.
        In the case of duplicate names, method selection is first in the order: xarray, earthkit, numpy.
        Otherwise it can be any function which can be called in the form `f(x, axis=axis, **kwargs)`
        to return the result of reducing an xp.ndarray over an integer valued axis
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    how_label : str
        Label to append to the name of the variable in the reduced object, default is nothing
    xp : T.Any
        The array namespace to use for the reduction. If None, it will be inferred from the dataarray.
    **kwargs :
        kwargs recognised by the how :func: `reduce`

    Returns
    -------
    A data array with reduce dimensions removed.

    """
    if xp is None:
        xp = _tools.array_namespace_from_object(dataarray)
    # If weighted, use xarray weighted methods
    if weights is not None:
        # Create any standard weights, e.g. latitude
        if isinstance(weights, str):
            _weights = _tools.standard_weights(dataarray, weights, **kwargs)
        else:
            _weights = weights
        # We ensure the callable is always a string
        if callable(how):
            how = weighted_how = how.__name__
        # map any alias methods:
        weighted_how = _tools.WEIGHTED_HOW_METHODS.get(how, how)

        red_array = dataarray.weighted(_weights).__getattribute__(weighted_how)(**kwargs)

    else:
        # If how is string, fetch function from dictionary:
        if isinstance(how, str) and how in dir(dataarray):
            red_array = dataarray.__getattribute__(how)(**kwargs)
        else:
            if isinstance(how, str):
                how = _tools.get_how_xp(how, xp=xp)
            assert callable(how), f"how method not recognised: {how}"

            red_array = dataarray.reduce(how, **kwargs)

    red_array = how_label_rename(red_array, how_label=how_label)

    if how_dropna:
        red_array = red_array.dropna(how_dropna)

    return red_array


def reduce(
    dataarray: xr.Dataset | xr.DataArray,
    *_args,
    **kwargs,
):
    """Reduce an xarray.dataarray or xarray.dataset using a specified `how` method.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data object to reduce
    how: str or callable
        Method used to reduce data. Default='mean', which will implement the xarray in-built mean.
        If string, it must be an in-built xarray reduce method, a earthkit how method or any numpy method.
        In the case of duplicate names, method selection is first in the order: xarray, earthkit, numpy.
        Otherwise it can be any function which can be called in the form `f(x, axis=axis, **kwargs)`
        to return the result of reducing an xp.ndarray over an integer valued axis
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_label : str
        Label to append to the name of the variable in the reduced object
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    **kwargs :
        kwargs recognised by the how :func: `reduce`

    Returns
    -------
    A data array with reduce dimensions removed.

    """
    # handle how as arg or kwarg
    kwargs["how"] = _args[0] if _args else kwargs.get("how", "mean")

    if isinstance(dataarray, xr.Dataset):
        out_ds = xr.Dataset().assign_attrs(dataarray.attrs)
        for var in dataarray.data_vars:
            out_da = _reduce_dataarray(dataarray[var], **kwargs)
            out_ds[out_da.name] = out_da
        return out_ds

    out = _reduce_dataarray(dataarray, **kwargs)
    return out


def rolling_reduce(dataarray: xr.Dataset | xr.DataArray, *_args, **kwargs) -> xr.Dataset | xr.DataArray:
    """Return reduced data using a moving window over which to apply the reduction.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data over which the moving window is applied according to the reduction method.
    windows :
        windows for the rolling groups, for example `time=10` to perform a reduction
        in the time dimension with a bin size of 10. the rolling groups can be defined
        over any number of dimensions. **see documentation for xarray.dataarray.rolling**.
    min_periods : integer
        The minimum number of observations in the window required to have a value
        (otherwise result is NaN). Default is to set **min_periods** equal to the size of the window.
        **see documentation for xarray.dataarray.rolling**
    center : bool
        Set the labels at the centre of the window, **see documentation for xarray.dataarray.rolling**.
    how_reduce : str,
        Function to be applied for reduction. Default is 'mean'.
    how_dropna : str
        Determine if dimension is removed from the output when we have at least one NaN or
        all NaN. **how_dropna** can be 'None', 'any' or 'all'. Default is 'any'.
    **kwargs :
        Any kwargs that are compatible with the select `how_reduce` method.

    Returns
    -------
    xr.DataArray or xr.Dataset (as provided)
    """
    if isinstance(dataarray, (xr.Dataset)):
        out_ds = xr.Dataset().assign_attrs(dataarray.attrs)
        for var in dataarray.data_vars:
            out_da = _rolling_reduce_dataarray(dataarray[var], *_args, **kwargs)
            out_ds[out_da.name] = out_da
        return out_ds
    else:
        return _rolling_reduce_dataarray(dataarray, *_args, **kwargs)


def _rolling_reduce_dataarray(
    dataarray: xr.DataArray, how_reduce="mean", how_dropna=None, chunk=True, **kwargs
) -> xr.DataArray:
    """Return reduced data using a moving window over which to apply the reduction.

    Parameters
    ----------
    dataarray : xr.DataArray
        Data over which the moving window is applied according to the reduction method.
    windows :
        windows for the rolling groups, for example `time=10` to perform a reduction
        in the time dimension with a bin size of 10. the rolling groups can be defined
        over any number of dimensions. **see documentation for xarray.dataarray.rolling**.
    min_periods : integer
        The minimum number of observations in the window required to have a value
        (otherwise result is NaN). Default is to set **min_periods** equal to the size of the window.
        **see documentation for xarray.dataarray.rolling**
    center : bool
        Set the labels at the centre of the window, **see documentation for xarray.dataarray.rolling**.
    how_reduce : str,
        Function to be applied for reduction. Default is 'mean'.
    how_dropna : str
        Determine if dimension is removed from the output when we have at least one NaN or
        all NaN. **how_dropna** can be 'None', 'any' or 'all'. Default is None.
    chunk: bool
        If True, the dataarray is chunked before the rolling operation.
    **kwargs :
        Any kwargs that are compatible with the select `how_reduce` method.

    Returns
    -------
    xr.DataArray
    """
    xp = _tools.array_namespace_from_object(dataarray)

    if chunk:
        dataarray = dataarray.chunk()
    # Expand dim kwarg to individual kwargs
    if isinstance(kwargs.get("dim"), dict):
        kwargs.update(kwargs.pop("dim"))

    window_dims = [str(_dim) for _dim in list(dataarray.dims) if _dim in list(kwargs)]
    rolling_kwargs_keys = ["min_periods", "center"] + window_dims
    rolling_kwargs_keys = [_kwarg for _kwarg in kwargs if _kwarg in rolling_kwargs_keys]
    rolling_kwargs = {_kwarg: kwargs.pop(_kwarg) for _kwarg in rolling_kwargs_keys}

    # Any kwargs left after above reductions are kwargs for reduction method
    reduce_kwargs = kwargs
    # Create rolling groups:
    data_rolling = dataarray.rolling(**rolling_kwargs)

    reduce_kwargs.setdefault("how", how_reduce)
    # TODO: remove type ignore when xarray puts types in stable location
    data_windowed = _reduce_dataarray(data_rolling, xp=xp, **reduce_kwargs)  # type: ignore

    data_windowed = _dropna(data_windowed, window_dims, how_dropna)

    data_windowed.attrs.update(dataarray.attrs)

    return data_windowed


def _dropna(data, dims, how):
    """Drop nan values from data."""
    if how in [None, "None", "none"]:
        return data

    for dim in dims:
        data = data.dropna(dim, how=how)
    return data


@_tools.time_dim_decorator
def resample(
    dataarray: xr.Dataset | xr.DataArray,
    frequency: str | int | float,
    time_dim: str = "time",
    how: str = "mean",
    skipna: bool = True,
    how_args: list[T.Any] = [],
    how_kwargs: dict[str, T.Any] = {},
    how_label: str | None = None,
    **kwargs,
) -> xr.Dataset | xr.DataArray:
    """Resample dataarray to a user-defined frequency using a user-defined "how" method.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray to be resampled.
    frequency : str, int, float
        The frequency at which to resample the chosen dimension. The format must be applicable
        to the chosen dimension.
    time_dim: str
        The dimension to resample along, default is `time`
    how: str
        The reduction method for resampling, default is `mean`
    how_label : str
        Label to append to the name of the variable in the reduced object, default is nothing
    skipna : bool
        If True, exclude missing values (na values) from the reduction.
    how_args : list
        List of arguments to be passed to the reduction method.
    how_kwargs : dict
        Dictionary of keyword arguments to be passed to the reduction method.
    **kwargs
        Keyword arguments to be passed to :func:`resample`. Defaults have been set as:
        `{"skipna": True}`

    Returns
    -------
    xr.Dataset | xr.DataArray
    """
    # Handle legacy API instances:
    time_dim = kwargs.pop("dim", time_dim)

    # Get any how kwargs into appropriate dictionary:
    for _k in ["q", "p"]:
        if _k in kwargs:
            how_kwargs[_k] = kwargs.pop(_k)
    # Translate and xarray frequencies to pandas language:
    frequency = _tools._PANDAS_FREQUENCIES_R.get(frequency, frequency)
    kwargs[time_dim] = frequency
    resample = dataarray.resample(skipna=skipna, **kwargs)
    result = resample.__getattribute__(how)(*how_args, dim=time_dim, **how_kwargs)

    result = how_label_rename(result, how_label=how_label)

    return result



# (C) Copyright 2024- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import numpy as np
import pytest
import xarray as xr
from earthkit.transforms.ensemble import (
    max,
    mean,
    min,
    reduce,
    std,
)


# Sample data for tests
@pytest.fixture
def sample_data():
    data = xr.DataArray(
        np.arange(12).reshape(3, 2, 2),
        dims=("ensemble_member", "lat", "lon"),
        coords={"ensemble_member": [0, 1, 2]},
    )
    return data


def test_reduce_raises_if_no_ensemble_dim_found():
    data = xr.DataArray(np.arange(8).reshape(2, 2, 2), dims=("foo", "lat", "lon"))
    with pytest.raises(ValueError, match="Unable to find dimension key for axis 'realization' in dataarray"):
        reduce(data)


def test_mean_actual_reduction(sample_data):
    # This tests integration if you want to use real data (no mocking)
    result = mean(sample_data, dim="ensemble_member")
    expected = sample_data.mean(dim="ensemble_member")
    assert np.allclose(result, expected)


def test_standard_deviation_actual(sample_data):
    result = std(sample_data, dim="ensemble_member")
    expected = sample_data.std(dim="ensemble_member")
    assert np.allclose(result, expected)


def test_min_max_actual(sample_data):
    result_min = min(sample_data, dim="ensemble_member")
    expected_min = sample_data.min(dim="ensemble_member")
    assert np.allclose(result_min, expected_min)

    result_max = max(sample_data, dim="ensemble_member")
    expected_max = sample_data.max(dim="ensemble_member")
    assert np.allclose(result_max, expected_max)



import functools
import importlib
import inspect
import logging
import types
import typing as T
from functools import wraps

import numpy as np
import pandas as pd
import xarray as xr
from earthkit.data import transform
from earthkit.data.utils.module_inputs_wrapper import _ensure_iterable, _ensure_tuple, signature_mapping
from earthkit.data.wrappers import Wrapper
from earthkit.utils.array import array_namespace

logger = logging.getLogger(__name__)

#: Mapping from pandas frequency strings to xarray time groups
_PANDAS_FREQUENCIES = {
    "D": "dayofyear",
    "W": "weekofyear",
    "M": "month",
    "ME": "month",
    "MS": "month",
    "H": "hour",
}
# Note this is not 100% reversible, 3 pandas freqs map to xarray's "month",
# but "month" will only map to "MS"
_PANDAS_FREQUENCIES_R = {v: k for k, v in _PANDAS_FREQUENCIES.items()}

#: The maximum limit of climatology time groups
_BIN_MAXES = {
    "hour": 24,
    "dayofyear": 366,
    "weekofyear": 53,
    "month": 12,
    "season": 4,
}


def ensure_list(thing) -> list[T.Any]:
    if isinstance(thing, list):
        return thing
    try:
        return thing.to_list()
    except AttributeError:
        return [thing]


def time_dim_decorator(func):
    @functools.wraps(func)
    def wrapper(
        dataarray: xr.Dataset | xr.DataArray,
        *args,
        time_dim: str | None = None,
        time_shift: dict | str | pd.Timedelta | None = None,
        remove_partial_periods: bool = False,
        **kwargs,
    ):
        if time_dim is None:
            try:
                time_dim = get_dim_key(dataarray, "t")
            except Exception:
                # Not able to find time dimension in object so let fail its own way
                func(dataarray, *args, **kwargs)

        if time_shift is not None:
            # Create timedelta from dict
            if isinstance(time_shift, dict):
                time_shift = pd.Timedelta(**time_shift)
            else:
                time_shift = pd.Timedelta(time_shift)

            # Convert timedelta to timedelta64 (TODO: may need to be more robust here)
            time_coord = dataarray.coords[time_dim] + time_shift  # type: ignore
            time_coord = time_coord.assign_attrs({"time_shift": f"{time_shift}"})

            dataarray = dataarray.assign_coords({time_dim: time_coord})

        result = func(dataarray, *args, time_dim=time_dim, **kwargs)

        # If we want only full days then chop off the first and last day if we have time_shift
        if remove_partial_periods and time_shift and isinstance(result, (xr.Dataset, xr.DataArray)):
            return result.isel({time_dim: slice(1, -1)})

        return result

    return wrapper


GROUPBY_KWARGS = ["frequency", "bin_widths"]


def groupby_kwargs_decorator(func):
    @functools.wraps(func)
    def wrapper(*args, groupby_kwargs: dict | None = None, **kwargs):
        groupby_kwargs = groupby_kwargs or {}
        new_kwargs = {}
        for k, v in kwargs.copy().items():
            if k in GROUPBY_KWARGS:
                groupby_kwargs.setdefault(k, v)
            else:
                new_kwargs[k] = v
        return func(*args, groupby_kwargs=groupby_kwargs, **new_kwargs)

    return wrapper


def season_order_decorator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if kwargs.get("frequency", "NOTseason") in ["season"]:
            result.reindex(season=["DJF", "MAM", "JJA", "SON"])
        return result

    return wrapper


def array_namespace_from_object(data_object: T.Any) -> types.ModuleType:
    """Attempt to infer the array namespace from the data object.

    Parameters
    ----------
    data_object : T.Any
        The data object from which to infer the array namespace.

    Returns
    -------
    types.ModuleType
        The inferred array namespace.

    Raises
    ------
    TypeError
        If the input data_object contains an compatible array interface,
        e.g. a xr.Dataset with mixed array namespaces.
    """
    if isinstance(data_object, xr.DataArray):
        return array_namespace(data_object.data)
    elif isinstance(data_object, xr.Dataset):
        data_vars = list(data_object.data_vars)
        xps = [array_namespace(data_object[var].data) for var in data_vars]
        if len(set(xps)) == 1:
            return xps[0]
        elif len(set(xps)) > 1:
            raise TypeError(
                "Data object contains variables with different array namespaces, "
                "cannot infer a single xp for computation."
            )
        else:
            raise TypeError("data_object must contain at least one data_variable to infer xp.")

    try:
        return array_namespace(data_object)
    except Exception:
        logger.warning(
            "Unable to infer array namespace from data_object, defaulting to numpy. "
            "If you are using a custom data object, please ensure it has a compatible array interface.",
        )
        return np


def nanaverage(data, weights=None, **kwargs):
    """Calculate the average of data ignoring nan-values.

    Parameters
    ----------
    data : array
        Data to average.
    weights:
        Weights to apply to the data for averaging.
        Weights will be normalised and must correspond to the
        shape of the data array and axis/axes that is/are
        averaged over.
    axis:
        axis/axes to compute the nanaverage over.
    kwargs:
        any other xp.nansum kwargs

    Returns
    -------
    Array mean of data (along axis) where nan-values are ignored
    and weights applied if provided.
    """
    xp = array_namespace_from_object(data)
    if weights is not None:
        # set weights to nan where data is nan:
        this_weights = xp.ones(data.shape) * weights
        this_weights[xp.isnan(data)] = xp.nan
        # Weights must be scaled to the sum of valid
        #  weights for each relevant axis:
        this_denom = xp.nansum(this_weights, **kwargs)
        # If averaging over an axis then we must add dummy
        # dimension[s] to the denominator to make compatible
        # with the weights.
        if kwargs.get("axis", None) is not None:
            reshape = list(this_weights.shape)
            reshape[kwargs.get("axis")] = 1
            this_denom = this_denom.reshape(reshape)

        # Scale weights to mean of valid weights:
        this_weights = this_weights / this_denom
        # Apply weights to data:
        _nanaverage = xp.nansum(data * this_weights, **kwargs)
    else:
        # If no weights, then nanmean will suffice
        _nanaverage = xp.nanmean(data, **kwargs)

    return _nanaverage


def standard_weights(dataarray: xr.DataArray, weights: str, **kwargs):
    """Implement any standard weights functions included in earthkit-transforms."""
    if weights in ["latitude", "lat"]:
        lat_weight_kwargs = {key: value for key, value in kwargs.items() if key in ["lat_key"]}
        return latitude_weights(dataarray, **lat_weight_kwargs)

    raise NotImplementedError(f"The selected weights method is not recognised or implemented yet: {weights}.")


def latitude_weights(dataarray: xr.DataArray, lat_key: str | None = None):
    """xarray.DataArray wrapper for latitude_weights.

    Detects the spatial dimensions latitude must be a coordinate of the dataarray.
    """
    if lat_key is None:
        lat_key = get_dim_key(dataarray, "y")

    lat_array = dataarray.coords.get(lat_key)
    if lat_array is not None:
        xp = array_namespace_from_object(lat_array[lat_key])
        return xp.cos(xp.radians(lat_array[lat_key]))

    raise KeyError(
        "Latitude variable name not detected or found in the dataarray. Please provide the correct key."
    )


HOW_METHODS = {
    "average": nanaverage,
    "nanaverage": nanaverage,
    "mean": np.nanmean,
    "stddev": np.nanstd,
    "std": np.nanstd,
    "stdev": np.nanstd,
    "sum": np.nansum,
    "max": np.nanmax,
    "min": np.nanmin,
    "median": np.nanmedian,
    "q": np.nanquantile,
    "quantile": np.nanquantile,
    "percentile": np.nanpercentile,
    "p": np.nanpercentile,
}

HOW_METHODS_MAPPING = {
    "average": "nanaverage",
    "mean": "nanmean",
    "stddev": "nanstd",
    "std": "nanstd",
    "stdev": "nanstd",
    "sum": "nansum",
    "max": "nanmax",
    "min": "nanmin",
    "median": "nanmedian",
    "q": "nanquantile",
    "quantile": "nanquantile",
    "percentile": "nanpercentile",
    "p": "nanpercentile",
}

WEIGHTED_HOW_METHODS = {
    "average": "mean",
    # "mean": "mean",
    "nanmean": "mean",
    "stddev": "std",
    # "std": "std",
    "stdev": "std",
    # "sum": "sum",
    # "sum_of_squares": "sum_of_squares",
    # "sum_of_weights": "sum_of_weights",
    "q": "quantile",
    # "quantile": "quantile",
    # "percentile": np.nanpercentile,
    # "p": np.nanpercentile,
}


# Libraries which are usable with reduce
ALLOWED_LIBS = {
    "numpy": "np",
    "cupy": "cp",
}

# Dictionary containing recognised weight functions.
WEIGHTS_DICT = {
    "latitude": latitude_weights,
    "lat": latitude_weights,
}


def get_how(how: str, how_methods=HOW_METHODS):
    try:
        how = how_methods[how]
    except KeyError:
        try:
            module, function = how.split(".")
        except Exception:
            raise ValueError(f"how method not recognised or found: how={how}")

        try:
            how = getattr(globals()[ALLOWED_LIBS[module]], function)
        except KeyError:
            raise ValueError(f"method must come from one of {ALLOWED_LIBS}")
        except AttributeError:
            raise AttributeError(f"module '{module}' has no attribute " f"'{function}'")

    return how


def resolve_function_from_path(path: str) -> T.Callable:
    """Given a string like 'numpy.mean' or 'some.module.func', dynamically import module and return callable.

    Parameters
    ----------
    path : str
        Fully qualified path to a function (e.g., 'numpy.mean').

    Returns
    -------
    T.Callable
        The resolved function object.


    Raises
    ------
    ValueError
        If the path is invalid or the function/module is not found.
    """
    if "." not in path:
        raise ValueError(f"Invalid path '{path}'. Must be in the form 'module.func'.")

    module_path, func_name = path.rsplit(".", 1)

    try:
        module = importlib.import_module(module_path)
    except ImportError as e:
        raise ValueError(f"Module '{module_path}' could not be imported: {e}") from e

    try:
        func = getattr(module, func_name)
    except AttributeError as e:
        raise ValueError(f"Function '{func_name}' not found in module '{module_path}'.") from e

    if not callable(func):
        raise ValueError(f"Resolved attribute '{func_name}' in '{module_path}' is not callable.")

    return func


def get_how_xp(
    how_str: str,
    xp: types.ModuleType | None = None,
    how_methods_mapping: dict[str, str] = HOW_METHODS_MAPPING,
    data_object: T.Any = None,
) -> T.Callable:
    """Resolve a method name to a callable from the given module (xp), using an optional mapping for aliases.

    Parameters
    ----------
    how_str : str
        The method name or alias.

    xp : module, optional
        The array API module (e.g., numpy). Defaults to numpy.

    how_methods_mapping : dict
        Mapping of aliases to method names.

    data_object : Any, optional
        The data object to infer the array API from, if xp is not provided explicitly

    Returns
    -------
    T.Callable
        The resolved method.

    Raises
    ------
    ValueError
        If the method cannot be found in xp.
    """
    # First check if the "how_str" has a `.`, if so, we assume it is a full path to the method
    if "." in how_str:
        return resolve_function_from_path(how_str)

    resolved_name = how_methods_mapping.get(how_str, how_str)

    # Check if the resolved name is in this script's globals
    if resolved_name in globals():
        return globals()[resolved_name]

    if xp is None:
        if data_object is not None:
            try:
                xp = array_namespace_from_object(data_object)
            except Exception:
                logger.warning("Unable to infer array namespace from data_object, defaulting to numpy.")
                xp = np
        else:
            # Default to numpy if no xp or data_object is provided
            xp = np

    for name in (resolved_name, how_str):
        if hasattr(xp, name):
            return getattr(xp, name)

    raise ValueError(f"how method not recognised or found: how={how_str} for xp={xp.__name__}.")


STANDARD_AXIS_KEYS: dict[str, list[str]] = {
    "y": ["lat", "latitude"],
    "x": ["lon", "long", "longitude"],
    "t": ["time", "valid_time", "forecast_reference_time"],
    "realization": ["ensemble_member", "ensemble", "member", "number", "realization", "realisation"],
}

STANDARD_AXIS_CF_NAMES: dict[str, list[str]] = {
    "y": ["projection_y_coordinate", "latitude", "grid_latitude"],
    "x": ["projection_x_coordinate", "longitude", "grid_longitude"],
    "t": ["time", "valid_time", "forecast_reference_time"],
    "realization": ["realization"],
}


def get_dim_key(
    dataarray: xr.Dataset | xr.DataArray,
    axis: str,
    raise_error: bool = False,
) -> str:
    """Return the key of the dimension.

    Parameters
    ----------
    dataarray : xr.Dataset or xr.DataArray
        The data to search for the dimension in.
    axis : str
        The axis to search for. This should be a CF standard axis key like 'x', 'y', 'z' or 't',
        or one of the earthkit-transforms recognised axis, e.g. "member".
    raise_error : bool
        If True, raises an error if the axis is not found in the dataarray.

    Returns
    -------
    str
        The key of the dimension in the dataarray that matches the axis.
    """
    # First check if the axis value is in any dim:
    for dim in dataarray.dims:
        if "axis" in dataarray[dim].attrs and dataarray[dim].attrs["axis"].lower() == axis.lower():
            return str(dim)

    # Then check if any dims have CF recognised standard names,
    #  Prioritised in order of the STANDARD_AXIS_CF_NAMES list order
    for standard_name in STANDARD_AXIS_CF_NAMES.get(axis.lower(), []):
        for dim in dataarray.dims:
            if dataarray[dim].attrs.get("standard_name") == standard_name:
                return str(dim)

    # Then check if any dims match our "standard" axis,
    #  Prioritised in order of the STANDARD_AXIS_KEYS list order
    for standard_axis_key in STANDARD_AXIS_KEYS.get(axis.lower(), []):
        if standard_axis_key in dataarray.dims:
            return standard_axis_key

    # We have not been able to detect, so raise an error
    if raise_error:
        raise ValueError(
            f"Unable to find dimension key for axis '{axis}' in dataarray with dims: {dataarray.dims}."
        )

    return axis


def get_spatial_info(dataarray, lat_key=None, lon_key=None):
    # Figure out the keys for the latitude and longitude variables
    if lat_key is None:
        lat_key = get_dim_key(dataarray, "y")
    if lon_key is None:
        lon_key = get_dim_key(dataarray, "x")

    # Get the geospatial dimensions of the data. In the case of regular data this
    #  will be 'lat' and 'lon'. For irregular data it could be any dimensions
    lat_dims = dataarray.coords[lat_key].dims
    lon_dims = dataarray.coords[lon_key].dims
    spatial_dims = [dim for dim in lat_dims] + [dim for dim in lon_dims if dim not in lat_dims]

    # Assert that latitude and longitude have the same dimensions
    #   (irregular data, e.g. x&y or obs)
    # or the dimensions are themselves (regular data, 'lat'&'lon')
    assert (lat_dims == lon_dims) or ((lat_dims == (lat_key,)) and (lon_dims) == (lon_key,))
    if lat_dims == lon_dims:
        regular = False
    elif (lat_dims == (lat_key,)) and (lon_dims) == (lon_key,):
        regular = True
    else:
        raise ValueError(
            "The geospatial dimensions have not not been correctly detected:\n"
            f"lat_key: {lat_key}; lat_dims: {lat_dims}\n"
            f"lon_key: {lon_key}; lon_dims: {lon_dims}\n"
        )
    spatial_info = {
        "lat_key": lat_key,
        "lon_key": lon_key,
        "regular": regular,
        "spatial_dims": spatial_dims,
    }
    return spatial_info


def _pandas_frequency_and_bins(
    frequency: str,
) -> tuple[str, int | None]:
    freq = frequency.lstrip("0123456789")
    bins = int(frequency[: -len(freq)]) or None
    freq = _PANDAS_FREQUENCIES.get(freq.lstrip(" "), frequency)
    return freq, bins


def groupby_time(
    dataarray: xr.Dataset | xr.DataArray,
    frequency: str | None = None,
    bin_widths: int | None = None,
    time_dim: str = "time",
):
    if frequency is None:
        try:
            frequency = xr.infer_freq(dataarray.time)
        except:  # noqa: E722
            raise ValueError(
                "Unable to infer time frequency from data; please pass the 'frequency' argument explicitly"
            )
        frequency, possible_bins = _pandas_frequency_and_bins(frequency)
        bin_widths = bin_widths or possible_bins

    if bin_widths is not None:
        grouped_data = groupby_bins(dataarray, frequency, bin_widths, time_dim=time_dim)
    else:
        try:
            grouped_data = dataarray.groupby(f"{time_dim}.{frequency}")
        except AttributeError:
            raise ValueError(
                f"Invalid frequency '{frequency}' - see xarray documentation for "
                f"a full list of valid frequencies."
            )

    return grouped_data


def groupby_bins(
    dataarray: xr.Dataset | xr.DataArray,
    frequency: str,
    bin_widths: list[int] | int = 1,
    time_dim: str = "time",
):
    if not isinstance(bin_widths, (list, tuple)):
        max_value = _BIN_MAXES[frequency]
        bin_widths = list(range(0, max_value + 1, bin_widths))
    try:
        grouped_data = dataarray.groupby_bins(f"{time_dim}.{frequency}", bin_widths)
    except AttributeError:
        raise ValueError(
            f"Invalid frequency '{frequency}' - see xarray documentation for "
            f"a full list of valid frequencies."
        )
    return grouped_data


def transform_inputs_decorator(
    kwarg_types: T.Dict[str, T.Any] = {},
    convert_types: T.Union[None, T.Tuple[T.Any], T.Dict[str, T.Tuple[T.Any]]] = None,
) -> T.Callable:
    """Transform the inputs to a function to match the requirements.

    Parameters
    ----------
    kwarg_types : Dict[str, type]
        Mapping of accepted object types for each arg/kwarg
    convert_types : Tuple[type] or Dict[str, Tuple[type]]
        Data types to try to convert. If a dict, applies per-argument.

    Returns
    -------
    Callable
        Wrapped function.
    """
    if convert_types is None:
        convert_types = {}

    def decorator(function: T.Callable) -> T.Callable:
        def _wrapper(_kwarg_types, _convert_types, *args, **kwargs):
            _kwarg_types = {**_kwarg_types}
            signature = inspect.signature(function)
            mapping = signature_mapping(signature, _kwarg_types)

            # Store positional arg names for extraction later
            arg_names = []
            for arg, name in zip(args, signature.parameters):
                arg_names.append(name)
                kwargs[name] = arg

            convert_kwargs = [k for k in kwargs if k in mapping]

            # Filter for convert_types
            if _convert_types:
                if not isinstance(_convert_types, dict):
                    _convert_types = {key: _convert_types for key in convert_kwargs}

                convert_kwargs = [
                    k
                    for k in convert_kwargs
                    if isinstance(kwargs[k], _ensure_tuple(_convert_types.get(k, ())))
                ]

            # Transform args/kwargs
            for key in convert_kwargs:
                value = kwargs[key]
                types_allowed = _ensure_iterable(mapping[key])
                if type(value) not in types_allowed:
                    for target_type in types_allowed:
                        try:
                            kwargs[key] = transform(value, target_type)
                        except Exception:
                            continue
                        break

            # Expand Wrapper objects
            for k, v in list(kwargs.items()):
                if isinstance(v, Wrapper):
                    try:
                        kwargs[k] = v.data
                    except Exception:
                        pass

            # Extract positional args again
            args = [kwargs.pop(name) for name in arg_names]
            return function(*args, **kwargs)

        @wraps(function)
        def wrapper(*args, _auto_inputs_transform=True, **kwargs):
            if not _auto_inputs_transform:
                return function(*args, **kwargs)
            return _wrapper(kwarg_types, convert_types, *args, **kwargs)

        return wrapper

    return decorator



"""Module for preservation of the legacy API, it will be removed in version 2.X of earthkit.transforms."""
import logging

from earthkit.transforms.spatial._aggregate import mask, reduce

logger = logging.getLogger(__name__)


def masks(
    *_args,
    **_kwargs,
):
    logger.warning(
        "earthkit.transforms.aggregate.spatial.masks is deprecated, "
        "please use earthkit.transforms.aggregate.spatial.mask instead."
    )
    return mask(*_args, **_kwargs)


__all__ = [mask, "masks", reduce]



import numpy as np
import pandas as pd
import pytest

# from earthkit.data.core.temporary import temp_directory
import xarray as xr
from earthkit import data as ek_data
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.transforms import temporal

# Use caching for speedy repeats
ek_data.settings.set("cache-policy", "user")


def get_data():
    remote_era5_file = earthkit_remote_test_data_file("era5_temperature_europe_2015.grib")
    return ek_data.from_source("url", remote_era5_file)


@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_temporal_reduce(in_data, expected_return_type):
    reduced_data = temporal.reduce(in_data, how="mean")
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" not in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == reduced_data.name
    else:
        assert "2t" in reduced_data
    reduced_data = temporal.reduce(in_data, how="mean", how_label="mean")
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" not in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t_mean" == reduced_data.name
    else:
        assert "2t_mean" in reduced_data


def test_standardise_time_basic():
    data = get_data().to_xarray()
    original_time = data.forecast_reference_time
    data_standardised = temporal.standardise_time(data)
    np.testing.assert_array_equal(original_time.values, data_standardised.forecast_reference_time.values)


def test_standardise_time_monthly():
    data = get_data().to_xarray()
    data_standardised = temporal.standardise_time(data, target_format="%Y-%m-15")
    assert all(
        pd.to_datetime(time_value).day == 15
        for time_value in data_standardised.forecast_reference_time.values
    )


@pytest.mark.parametrize(
    "how",
    ("mean", "median", "min", "max", "std", "sum"),
)
def test_temporal_reduce_hows(how):
    in_data = get_data().to_xarray()
    expected_return_type = xr.Dataset
    reduced_data = temporal.reduce(in_data, how=how)
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" not in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == reduced_data.name
    else:
        assert "2t" in reduced_data

    reduced_data = temporal.reduce(in_data, how=how, how_label=how)
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" not in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert f"2t_{how}" == reduced_data.name
    else:
        assert f"2t_{how}" in reduced_data


@pytest.mark.parametrize(
    "method",
    ("mean", "median", "min", "max", "std", "sum"),
)
@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_temporal_methods(method, in_data, expected_return_type):
    reduced_data = temporal.__getattribute__(method)(in_data)
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" not in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == reduced_data.name
    else:
        assert "2t" in reduced_data


@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        # [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_temporal_daily_reduce_intypes(in_data, expected_return_type, how="mean"):
    reduced_data = temporal.daily_reduce(in_data, how=how)
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == reduced_data.name
    else:
        assert "2t" in reduced_data


@pytest.mark.parametrize(
    "how",
    ("mean", "median", "min", "max", "std", "sum"),
)
def test_temporal_daily_reduce_hows(how, in_data=get_data().to_xarray(), expected_return_type=xr.Dataset):
    reduced_data = temporal.daily_reduce(in_data, how=how)
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == reduced_data.name
    else:
        assert "2t" in reduced_data


@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        # [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_temporal_monthly_reduce_intypes(in_data, expected_return_type, how="mean"):
    reduced_data = temporal.monthly_reduce(in_data, how=how)
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == reduced_data.name
    else:
        assert "2t" in reduced_data


@pytest.mark.parametrize(
    "how",
    ("mean", "median", "min", "max", "std", "sum"),
)
def test_temporal_monthly_reduce_hows(how, in_data=get_data().to_xarray(), expected_return_type=xr.Dataset):
    reduced_data = temporal.monthly_reduce(in_data, how=how)
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == reduced_data.name
    else:
        assert "2t" in reduced_data


@pytest.mark.parametrize(
    "method",
    (
        "daily_mean",
        "daily_median",
        "daily_min",
        "daily_max",
        "daily_std",
        "daily_sum",
        "monthly_mean",
        "monthly_median",
        "monthly_min",
        "monthly_max",
        "monthly_std",
        "monthly_sum",
    ),
)
@pytest.mark.parametrize(
    "in_data, expected_return_type",
    (
        [get_data(), xr.Dataset],
        [get_data().to_xarray(), xr.Dataset],
        # [get_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_temporal_daily_monthly_methods(method, in_data, expected_return_type):
    reduced_data = temporal.__getattribute__(method)(in_data)
    assert isinstance(reduced_data, expected_return_type)
    assert "forecast_reference_time" in list(reduced_data.dims)
    if expected_return_type == xr.DataArray:
        assert "2t" == reduced_data.name
    else:
        assert "2t" in reduced_data



import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import xarray as xr

# from earthkit.data.core.temporary import temp_directory
from earthkit import data as ek_data
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.transforms import spatial
from earthkit.transforms.spatial import _aggregate as _spatial
from shapely.geometry import Polygon

try:
    import rasterio  # noqa: F401

    rasterio_available = True
except ImportError:
    rasterio_available = False

# Use caching for speedy repeats
ek_data.settings.set("cache-policy", "user")


SAMPLE_ARRAY = xr.DataArray(
    [
        [1, 1, 1, 1],
        [2, 2, 2, 2],
        [3, 3, 3, 3],
    ],
    dims=["latitude", "longitude"],
    coords={
        "latitude": [0, 60, 90],  # Chosen for latitude weight tests
        "longitude": [0, 30, 60, 90],
    },
)


class dummy_class:
    def __init__(self):
        self.to_pandas = pd.DataFrame
        self.to_geopandas = pd.DataFrame


def get_grid_data():
    remote_era5_file = earthkit_remote_test_data_file("era5_temperature_europe_20150101.grib")
    return ek_data.from_source("url", remote_era5_file)


def get_shape_data():
    if rasterio_available:
        remote_nuts_url = earthkit_remote_test_data_file("NUTS_RG_60M_2021_4326_LEVL_0.geojson")
        return ek_data.from_source("url", remote_nuts_url)
    return dummy_class()


@pytest.mark.skipif(
    not rasterio_available,
    reason="rasterio is not available",
)
def test_spatial_mask():
    single_masked_data = spatial.mask(get_grid_data(), get_shape_data(), union_geometries=True)
    assert isinstance(single_masked_data, xr.Dataset)


@pytest.mark.skipif(
    not rasterio_available,
    reason="rasterio is not available",
)
@pytest.mark.parametrize(
    "era5_data, nuts_data, expected_result_type",
    (
        [get_grid_data(), get_shape_data(), xr.Dataset],
        [get_grid_data().to_xarray(), get_shape_data(), xr.Dataset],
        [get_grid_data().to_xarray(), get_shape_data().to_pandas(), xr.Dataset],
        [get_grid_data(), get_shape_data().to_pandas(), xr.Dataset],
        [get_grid_data().to_xarray()["2t"], get_shape_data(), xr.DataArray],
    ),
)
def test_spatial_masks_with_ek_objects(era5_data, nuts_data, expected_result_type):
    masked_data = spatial.mask(era5_data, nuts_data)
    assert isinstance(masked_data, expected_result_type)
    assert "index" in masked_data.dims
    assert len(masked_data["index"]) == len(nuts_data)


@pytest.mark.parametrize(
    "era5_data, expected_result_type",
    (
        [get_grid_data(), xr.Dataset],
        [get_grid_data().to_xarray(), xr.Dataset],
        [get_grid_data().to_xarray()["2t"], xr.DataArray],
    ),
)
def test_spatial_reduce_no_geometry(era5_data, expected_result_type):
    reduced_data = spatial.reduce(era5_data)

    assert isinstance(reduced_data, expected_result_type)
    assert list(reduced_data.dims) == ["forecast_reference_time"]


def test_spatial_reduce_no_geometry_result():
    reduced_data = spatial.reduce(SAMPLE_ARRAY, how="mean")
    assert reduced_data.values == 2.0
    reduced_data = spatial.reduce(SAMPLE_ARRAY, how="mean", weights="latitude")
    assert np.isclose(reduced_data.values, 1 + (1.0 / 3))


@pytest.mark.skipif(
    not rasterio_available,
    reason="rasterio is not available",
)
@pytest.mark.parametrize(
    "era5_data, nuts_data, expected_result_type",
    (
        [get_grid_data(), get_shape_data(), xr.Dataset],
        [get_grid_data().to_xarray(), get_shape_data(), xr.Dataset],
        [get_grid_data().to_xarray(), get_shape_data().to_pandas(), xr.Dataset],
        [get_grid_data(), get_shape_data().to_pandas(), xr.Dataset],
        [get_grid_data().to_xarray()["2t"], get_shape_data(), xr.DataArray],
    ),
)
def test_spatial_reduce_with_geometry(era5_data, nuts_data, expected_result_type):
    reduced_data = spatial.reduce(era5_data, nuts_data)
    assert isinstance(reduced_data, expected_result_type)
    assert all([dim in ["forecast_reference_time", "index"] for dim in reduced_data.dims])
    assert len(reduced_data["index"]) == len(nuts_data)


@pytest.mark.skipif(
    not rasterio_available,
    reason="rasterio is not available",
)
def test_spatial_reduce_with_precomputed_mask():
    era5_data_xr = get_grid_data().to_xarray()["2t"]
    ones = (era5_data_xr.isel(forecast_reference_time=0) * 0 + 1).astype(int).rename("mask")
    mask = spatial.mask(ones, get_shape_data(), all_touched=False)
    mask_arrays = [mask.sel(index=index) for index in mask.index]
    reduced_data_test = spatial.reduce(era5_data_xr, geodataframe=get_shape_data())

    # reduce with a single mask
    reduced_data = spatial.reduce(era5_data_xr, mask_arrays=mask_arrays[0])
    assert isinstance(reduced_data, xr.DataArray)
    assert all([dim in ["forecast_reference_time", "index"] for dim in reduced_data.dims])
    assert reduced_data.equals(reduced_data_test.isel(index=0))

    # reduce with list of masks
    reduced_data = spatial.reduce(era5_data_xr, mask_arrays=mask_arrays)
    assert isinstance(reduced_data, xr.DataArray)
    assert all([dim in ["forecast_reference_time", "index"] for dim in reduced_data.dims])
    assert len(reduced_data["index"]) == len(mask_arrays)
    assert reduced_data.equals(reduced_data_test)


@pytest.mark.skipif(
    not rasterio_available,
    reason="rasterio is not available",
)
@pytest.mark.parametrize(
    "era5_data, nuts_data, expected_result_type",
    ([get_grid_data().to_xarray(), get_shape_data().to_pandas(), xr.Dataset],),
)
def test_spatial_reduce_with_geometry_and_latitude_weights(era5_data, nuts_data, expected_result_type):
    reduced_data = spatial.reduce(era5_data, nuts_data, weights="latitude")
    assert isinstance(reduced_data, expected_result_type)
    assert all([dim in ["forecast_reference_time", "index"] for dim in reduced_data.dims])
    assert len(reduced_data["index"]) == len(nuts_data)

    # Ensure weights works with an abstract name for latitude
    era5_data = era5_data.rename({"latitude": "elephant"})
    reduced_data = spatial.reduce(era5_data, nuts_data, weights="latitude", lat_key="elephant")
    assert isinstance(reduced_data, expected_result_type)
    assert all([dim in ["forecast_reference_time", "index"] for dim in reduced_data.dims])
    assert len(reduced_data["index"]) == len(nuts_data)


@pytest.mark.skipif(
    not rasterio_available,
    reason="rasterio is not available",
)
def test_mask_kwargs():
    era5_data = get_grid_data()
    era5_xr = era5_data.to_xarray()

    nuts_data = get_shape_data()
    nuts_gpd = nuts_data.to_geopandas()
    nuts_DK = nuts_gpd[nuts_gpd["CNTR_CODE"] == "DK"]

    masked_data = spatial.mask(era5_xr, nuts_DK, all_touched=True)
    assert len(np.where(~np.isnan(masked_data["2t"].values.flat))[0]) == 3432

    masked_data = spatial.mask(era5_xr, nuts_DK, all_touched=False)
    assert len(np.where(~np.isnan(masked_data["2t"].values.flat))[0]) == 2448

    reduced_data = spatial.reduce(era5_xr, nuts_DK, all_touched=True)
    reduced_data_nested = spatial.reduce(era5_xr, nuts_DK, mask_kwargs=dict(all_touched=True))
    xr.testing.assert_equal(reduced_data, reduced_data_nested)
    np.testing.assert_allclose(reduced_data["2t"].mean(), 279.4813)

    reduced_data_2 = spatial.reduce(era5_xr, nuts_DK, all_touched=False)
    reduced_data_nested_2 = spatial.reduce(era5_xr, nuts_DK, mask_kwargs=dict(all_touched=False))
    xr.testing.assert_equal(reduced_data_2, reduced_data_nested_2)
    np.testing.assert_allclose(reduced_data_2["2t"].mean(), 279.54733)


def create_test_dataarray():
    lat = np.linspace(-90, 90, 10)
    lon = np.linspace(-180, 180, 20)
    data = np.random.rand(10, 20)
    return xr.DataArray(
        data,
        coords={"lat": lat, "lon": lon},
        dims=("lat", "lon"),
        name="test_var",
    )


def create_test_geodataframe():
    polygons = [Polygon([(-180, -90), (-180, 90), (180, 90), (180, -90)])]
    return gpd.GeoDataFrame(geometry=polygons, index=[1])


def test_reduce_mean():
    dataarray = create_test_dataarray()
    result = _spatial._reduce_dataarray_as_xarray(dataarray, how="mean")
    assert isinstance(result, xr.DataArray)
    assert "lat" not in result.dims
    assert "lon" not in result.dims


def test_reduce_with_geodataframe():
    dataarray = create_test_dataarray()
    geodataframe = create_test_geodataframe()
    result = _spatial._reduce_dataarray_as_xarray(dataarray, geodataframe=geodataframe, how="mean")
    assert isinstance(result, xr.DataArray)
    assert "index" in result.dims  # Default mask_dim is "index"


def test_reduce_with_weights():
    dataarray = create_test_dataarray()
    result = _spatial._reduce_dataarray_as_xarray(dataarray, how="mean", weights="latitude")
    assert isinstance(result, xr.DataArray)


def test_reduce_invalid_how():
    dataarray = create_test_dataarray()
    with pytest.raises(ValueError):
        _spatial._reduce_dataarray_as_xarray(dataarray, how="invalid_method")


def test_reduce_with_mask():
    dataarray = create_test_dataarray()
    mask = xr.DataArray(
        np.random.randint(0, 2, size=dataarray.shape), coords=dataarray.coords, dims=dataarray.dims
    )
    result = _spatial._reduce_dataarray_as_xarray(dataarray, mask_arrays=[mask], how="sum")
    assert isinstance(result, xr.DataArray)


def test_return_geometry_as_coord():
    dataarray = create_test_dataarray()
    geodataframe = create_test_geodataframe()
    result = _spatial._reduce_dataarray_as_xarray(
        dataarray, geodataframe=geodataframe, return_geometry_as_coord=True
    )
    assert "geometry" in result.coords
    assert len(result.coords["geometry"].values) == len(geodataframe)


def test_reduce_as_pandas():
    dataarray = create_test_dataarray()
    result = _spatial._reduce_dataarray_as_pandas(dataarray, how="mean")
    assert isinstance(result, pd.DataFrame)


def test_reduce_as_pandas_with_geodataframe():
    dataarray = create_test_dataarray()
    geodataframe = create_test_geodataframe()
    result = _spatial._reduce_dataarray_as_pandas(dataarray, geodataframe=geodataframe, how="mean")
    assert isinstance(result, pd.DataFrame)
    assert not result.empty


def test_reduce_as_pandas_compact():
    dataarray = create_test_dataarray()
    geodataframe = create_test_geodataframe()
    result = _spatial._reduce_dataarray_as_pandas(
        dataarray, geodataframe=geodataframe, compact=True, how="mean"
    )
    assert isinstance(result, pd.DataFrame)
    assert f"{dataarray.name}" in result.columns


def test_reduce_as_pandas_without_geodataframe():
    dataarray = create_test_dataarray()
    result = _spatial._reduce_dataarray_as_pandas(dataarray, how="sum")
    assert isinstance(result, pd.DataFrame)
    assert not result.empty



import logging
import typing as T

import numpy as np
import pandas as pd
import xarray as xr
from earthkit.transforms import _tools
from earthkit.transforms._aggregate import how_label_rename, resample
from earthkit.transforms._aggregate import reduce as _reduce
from earthkit.transforms._aggregate import rolling_reduce as _rolling_reduce

logger = logging.getLogger(__name__)


@_tools.time_dim_decorator
@_tools.transform_inputs_decorator()
def standardise_time(
    dataarray: xr.Dataset | xr.DataArray,
    target_format: str = "%Y-%m-%d %H:%M:%S",
    time_dim: str | None = None,
) -> xr.Dataset | xr.DataArray:
    """Convert time coordinates to a standard format using the Gregorian calendar.

    This function is helpful when combining data from different sources with
    different time standards or calendars - for example, when combining data
    which uses a 360-day calendar with data which uses a Gregorian calendar. All
    data passed into this function will be converted to a standard Gregorian
    calendar format.

    Parameters
    ----------
    dataarray : xr.Dataset or xr.DataArray
        Data object with a time coordinate to be standardised.
    target_format : str, optional
        Datetime format to use when creating the standardised datetime object.
        This can be used to change the resolution of the datetime object - for
        example, "%Y-%m-%d" will reduce to daily resolution - or to fix elements
        of the datetime object - for example, "%Y-%m-15" would reduce to monthly
        resolution and fix the date to the 15th of each month.
    time_dim : str, optional
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
        If you do not want to aggregate along the time dimension use earthkit.transforms.aggregate.reduce

    Returns
    -------
    xr.Dataset | xr.DataArray
        Data object with the time coordinate standardised to the specified format

    """
    try:
        source_times = [time_value.strftime(target_format) for time_value in dataarray[time_dim].data]
    except AttributeError:
        source_times = [
            pd.to_datetime(time_value).strftime(target_format) for time_value in dataarray[time_dim].data
        ]

    standardised_times = np.array(
        [pd.to_datetime(time_string).to_datetime64() for time_string in source_times]
    )

    dataarray = dataarray.assign_coords({time_dim: standardised_times})

    history = dataarray.attrs.get("history", "")
    history += (
        "The time coordinate of this data has been standardised with "
        "earthkit.transforms.aggregate.temporal.standardise_time."
    )
    dataarray = dataarray.assign_attrs({"history": history})

    return dataarray


@_tools.transform_inputs_decorator()
@_tools.time_dim_decorator
def reduce(
    dataarray: xr.Dataset | xr.DataArray,
    *_args,
    time_dim: str | None = None,
    **kwargs,
) -> xr.Dataset | xr.DataArray:
    """Reduce an xarray.dataarray/dataset along the time/date dimension using a specified `how` method.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data object to reduce
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
        If you do not want to aggregate along the time dimension use earthkit.transforms.aggregate.reduce
    how: str or callable
        Method used to reduce data. Default='mean', which will implement the xarray in-built mean.
        If string, it must be an in-built xarray reduce method, a earthkit how method or any numpy method.
        In the case of duplicate names, method selection is first in the order: xarray, earthkit, numpy.
        Otherwise it can be any function which can be called in the form `f(x, axis=axis, **kwargs)`
        to return the result of reducing an array over an integer valued axis
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _{how}
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    **kwargs :
        kwargs recognised by the how :func: `earthkit.transforms.aggregate.reduce`

    Returns
    -------
    xr.Dataset | xr.DataArray
        A dataarray reduced in the time dimension using the specified method

    """
    if "frequency" in kwargs:
        return resample(dataarray, *_args, time_dim=time_dim, **kwargs)

    reduce_dims = _tools.ensure_list(kwargs.get("dim", []))
    if time_dim is not None and time_dim not in reduce_dims:
        reduce_dims.append(time_dim)
    kwargs["dim"] = reduce_dims

    return _reduce(dataarray, *_args, **kwargs)


def mean(
    *_args,
    **kwargs,
):
    """Calculate the mean of an xarray.dataarray or xarray.dataset along the time/date dimension.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data object to reduce
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
        If you do not want to aggregate along the time dimension use earthkit.transforms.aggregate.reduce
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _mean
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    **kwargs :
        kwargs recognised by the how :func: `earthkit.transforms.aggregate.reduce`

    Returns
    -------
    xr.Dataset | xr.DataArray
        A dataarray of the mean value in the time dimensions

    """
    kwargs["how"] = "mean"
    return reduce(*_args, **kwargs)


def median(
    *_args,
    **kwargs,
):
    """Calculate the median of an xarray.dataarray or xarray.dataset along the time/date dimension.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data object to reduce
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
        If you do not want to aggregate along the time dimension use earthkit.transforms.aggregate.reduce
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _mean
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    **kwargs :
        kwargs recognised by the how :func: `earthkit.transforms.aggregate.reduce`

    Returns
    -------
    xr.Dataset | xr.DataArray
        A dataarray of the median value in the time dimensions

    """
    kwargs["how"] = "median"
    return reduce(*_args, **kwargs)


def min(
    *_args,
    **kwargs,
):
    """Calculate the mn of an xarray.dataarray or xarray.dataset along the time/date dimension.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data object to reduce
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
        If you do not want to aggregate along the time dimension use earthkit.transforms.aggregate.reduce
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _mean
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    **kwargs :
        kwargs recognised by the how :func: `earthkit.transforms.aggregate.reduce`

    Returns
    -------
    xr.Dataset | xr.DataArray
        A dataarray of the minimum value in the time dimensions

    """
    kwargs["how"] = "min"
    return reduce(*_args, **kwargs)


def max(
    *_args,
    **kwargs,
):
    """Calculate the max of an xarray.dataarray or xarray.dataset along the time/date dimension.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data object to reduce
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
        If you do not want to aggregate along the time dimension use earthkit.transforms.aggregate.reduce
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _mean
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    **kwargs :
        kwargs recognised by the how :func: `earthkit.transforms.aggregate.reduce`

    Returns
    -------
    xr.Dataset | xr.DataArray
        A dataarray of the maximum value in the time dimensions

    """
    kwargs["how"] = "max"
    return reduce(*_args, **kwargs)


def std(
    *_args,
    **kwargs,
) -> xr.Dataset | xr.DataArray:
    """Calculate the standard deviation of an xarray.dataarray/dataset along the time/date dimension.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data object to reduce
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
        If you do not want to aggregate along the time dimension use earthkit.transforms.aggregate.reduce
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _mean
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    **kwargs :
        kwargs recognised by the how :func: `earthkit.transforms.aggregate.reduce`

    Returns
    -------
    xr.Dataset | xr.DataArray
        A dataarray of the standard deviation in the time dimensions

    """
    kwargs["how"] = "std"
    return reduce(*_args, **kwargs)


def sum(
    *_args,
    **kwargs,
) -> xr.Dataset | xr.DataArray:
    """Calculate the standard deviation of an xarray.dataarray/dataset along the time/date dimension.

    With the option to apply weights either directly or using a specified
    `weights` method.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data object to reduce
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
        If you do not want to aggregate along the time dimension use earthkit.transforms.aggregate.reduce
    weights : str
        Choose a recognised method to apply weighting. Currently available methods are; 'latitude'
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _mean
    how_dropna : str
        Choose how to drop nan values.
        Default is None and na values are preserved. Options are 'any' and 'all'.
    **kwargs :
        kwargs recognised by the how :func: `earthkit.transforms.aggregate.reduce`

    Returns
    -------
    xr.Dataset | xr.DataArray
        A dataarray summed in the time dimensions


    """
    kwargs["how"] = "sum"
    return reduce(*_args, **kwargs)


@_tools.transform_inputs_decorator()
@_tools.time_dim_decorator
def daily_reduce(
    dataarray: xr.Dataset | xr.DataArray,
    how: str | T.Callable = "mean",
    time_dim: str | None = None,
    **kwargs,
) -> xr.Dataset | xr.DataArray:
    """Group data by day and reduce using the given how method.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    how: str or callable
        Method used to reduce data. Default='mean', which will implement the xarray in-built mean.
        If string, it must be an in-built xarray reduce method, a earthkit how method or any numpy method.
        In the case of duplicate names, method selection is first in the order: xarray, earthkit, numpy.
        Otherwise it can be any function which can be called in the form `f(x, axis=axis, **kwargs)`
        to return the result of reducing an array over an integer valued axis
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _daily_{how}
    **kwargs
        Keyword arguments to be passed to :func:`reduce`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to daily values using the specified method
    """
    # If time_dim in dimensions then use resample, this should be faster.
    #  At present, performance differences are small, but resampling can be improved by handling as
    #  a pandas dataframes. resample function should be updated to do this.
    #  NOTE: force_groupby is an intentionally undocumented kwarg for debug purposes
    if time_dim in dataarray.dims and not kwargs.pop("force_groupby", False):
        kwargs.setdefault("frequency", "D")
        red_array = resample(dataarray, time_dim=time_dim, how=how, **kwargs)
    else:
        # Otherwise, we groupby, with specifics set up for daily and handling both datetimes and timedeltas
        if dataarray[time_dim].dtype in ["<M8[ns]"]:  # datetime
            group_key = "date"
        elif dataarray[time_dim].dtype in ["<m8[ns]"]:  # timedelta
            group_key = "days"
        else:
            raise TypeError(f"Invalid type for time dimension ({time_dim}): {dataarray[time_dim].dtype}")

        grouped_data = _tools.groupby_time(dataarray, time_dim=time_dim, frequency=group_key)
        # If how is string and inbuilt method of grouped_data, we apply
        if isinstance(how, str) and how in dir(grouped_data):
            red_array = grouped_data.__getattribute__(how)(**kwargs)
        else:
            # If how is string, fetch function from dictionary:
            if isinstance(how, str):
                how = _tools.get_how_xp(how, data_object=dataarray)
            assert callable(how), f"how method not recognised: {how}"

            red_array = grouped_data.reduce(how, **kwargs)
        try:
            red_array[group_key] = pd.DatetimeIndex(red_array[group_key].data)
        except TypeError:
            logger.warning(f"Failed to convert {group_key} to datetime, it may already be a datetime object")

        red_array = how_label_rename(red_array, kwargs.get("how_label"))

    return red_array


def daily_mean(*_args, **kwargs) -> xr.Dataset | xr.DataArray:
    """Return the daily mean of the datacube.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`reduce`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to daily mean values
    """
    return daily_reduce(*_args, how="mean", **kwargs)


def daily_median(*_args, **kwargs):
    """Return the daily median of the datacube.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`reduce`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to daily median values
    """
    return daily_reduce(*_args, how="median", **kwargs)


def daily_max(*_args, **kwargs):
    """Calculate the daily maximum.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`reduce`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to daily max values
    """
    return daily_reduce(*_args, how="max", **kwargs)


def daily_min(*_args, **kwargs):
    """Calculate the daily minimum.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`reduce`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to daily min values
    """
    return daily_reduce(*_args, how="min", **kwargs)


def daily_std(*_args, **kwargs):
    """Calculate the daily standard deviation.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`reduce`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to daily standard deviation values
    """
    return daily_reduce(*_args, how="std", **kwargs)


def daily_sum(*_args, **kwargs):
    """Calculate the daily sum (accumulation).

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`reduce`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to daily sum values
    """
    return daily_reduce(*_args, how="sum", **kwargs)


@_tools.transform_inputs_decorator()
@_tools.time_dim_decorator
def monthly_reduce(
    dataarray: xr.Dataset | xr.DataArray,
    how: str | T.Callable = "mean",
    time_dim: str | None = None,
    **kwargs,
):
    """Group data by day and reduce using the given how method.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    how: str or callable
        Method used to reduce data. Default='mean', which will implement the xarray in-built mean.
        If string, it must be an in-built xarray reduce method, a earthkit how method or any numpy method.
        In the case of duplicate names, method selection is first in the order: xarray, earthkit, numpy.
        Otherwise it can be any function which can be called in the form `f(x, axis=axis, **kwargs)`
        to return the result of reducing an array over an integer valued axis
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    how_label : str
        Label to append to the name of the variable in the reduced object, default is _monthly_{how}

    **kwargs
        Keyword arguments to be passed to :func:`reduce`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to monthly values using the specified method
    """
    # If time_dim in dimensions then use resample, this should be faster.
    #  At present, performance differences are small, but resampling can be improved by handling as
    #  a pandas dataframes. reample function should be updated to do this.
    #  NOTE: force_groupby is an undocumented kwarg for debug purposes
    if time_dim in dataarray.dims and not kwargs.pop("force_groupby", False):
        kwargs.setdefault("frequency", "MS")
        red_array = resample(dataarray, time_dim=time_dim, how=how, **kwargs)
    else:
        # Otherwise, we groupby, with specifics set up for monthly and handling both datetimes and timedeltas
        if dataarray[time_dim].dtype in ["<M8[ns]"]:  # datetime
            # create a year-month coordinate
            years = dataarray[f"{time_dim}.year"]
            months = dataarray[f"{time_dim}.month"]
            dataarray.coords["year_months"] = years * 100 + months
            # Group by this:
            grouped_data = dataarray.groupby("year_months")
        elif dataarray[time_dim].dtype in ["<m8[ns]"]:  # timedelta
            raise TypeError(
                "Monthly aggregation is not support for timedelta objects. "
                "Please choose an alternative coordinate, or convert dimension to datatime object"
            )
        else:
            raise TypeError(f"Invalid type for time dimension ({time_dim}): {dataarray[time_dim].dtype}")

        # If how is string and inbuilt method of grouped_data, we apply
        if isinstance(how, str) and how in dir(grouped_data):
            red_array = grouped_data.__getattribute__(how)(**kwargs)
        else:
            # If how is string, fetch function from dictionary:
            if isinstance(how, str):
                how = _tools.get_how_xp(how, data_object=dataarray)
            assert callable(how), f"how method not recognised: {how}"

            red_array = grouped_data.reduce(how, **kwargs)
        # Remove the year_months coordinate
        del red_array["year_months"]

        red_array = how_label_rename(red_array, kwargs.get("how_label"))

    return red_array


def monthly_mean(
    *_args,
    **kwargs,
):
    """Calculate the monthly mean.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`resample`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to monthly mean values
    """
    return monthly_reduce(*_args, how="mean", **kwargs)


def monthly_median(
    *_args,
    **kwargs,
):
    """Calculate the monthly median.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`resample`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to monthly median values
    """
    return monthly_reduce(*_args, how="median", **kwargs)


def monthly_min(
    *_args,
    **kwargs,
):
    """Calculate the monthly min.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`resample`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to monthly minimum values
    """
    return monthly_reduce(*_args, how="min", **kwargs)


def monthly_max(
    *_args,
    **kwargs,
):
    """Calculate the monthly max.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`resample`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to monthly maximum values
    """
    return monthly_reduce(*_args, how="max", **kwargs)


def monthly_std(
    *_args,
    **kwargs,
):
    """Calculate the monthly standard deviation.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`resample`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to monthly standard deviation values
    """
    return monthly_reduce(*_args, how="std", **kwargs)


def monthly_sum(
    *_args,
    **kwargs,
):
    """Calculate the monthly sum/accumulation along the time dimension.

    Parameters
    ----------
    dataarray : xr.DataArray
        DataArray containing a `time` dimension.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object to use for the calculation,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    time_shift : (optional) None, timedelta or dict
        A time shift to apply to the data prior to calculation, e.g. to change the local time zone.
        It can be provided as any object that can be understood by `pandas.Timedelta`, a dictionary is passed
        as kwargs to `pandas.Timedelta`. Default is None.
    remove_partial_periods : bool
        If True and a time_shift has been applied, the first and last time steps are removed to ensure
        equality in sampling periods. Default is False.
    **kwargs
        Keyword arguments to be passed to :func:`resample`.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced to monthly sum values
    """
    return monthly_reduce(*_args, how="sum", **kwargs)


@_tools.transform_inputs_decorator()
@_tools.time_dim_decorator
def rolling_reduce(
    dataarray: xr.Dataset | xr.DataArray,
    window_length: int | None = None,
    time_dim: str | None = None,
    **kwargs,
):
    """Return reduced data using a moving window over the time dimension.

    Parameters
    ----------
    dataarray : xr.DataArray or xr.Dataset
        Data over which the moving window is applied according to the reduction method.
    window_length :
        Length of window for the rolling groups along the time dimension.
        **see documentation for xarray.dataarray.rolling**.
    time_dim : str
        Name of the time dimension, or coordinate, in the xarray object,
        default behaviour is to deduce time dimension from
        attributes of coordinates, then fall back to `"time"`.
    min_periods : integer
        The minimum number of observations in the window required to have a value
        (otherwise result is NaN). Default is to set **min_periods** equal to the size of the window.
        **see documentation for xarray.dataarray.rolling**
    center : bool
        Set the labels at the centre of the window, **see documentation for xarray.dataarray.rolling**.
    how_reduce : str,
        Function to be applied for reduction. Default is 'mean'.
    how_dropna : str
        Determine if dimension is removed from the output when we have at least one NaN or
        all NaN. **how_dropna** can be 'None', 'any' or 'all'. Default is 'any'.
    windows : dict[str, int]
        Any other windows to apply to other dimensions in the dataset/dataarray
    **kwargs :
        Any kwargs that are compatible with the select `how_reduce` method.

    Returns
    -------
    xr.DataArray | xr.Dataset
        A dataarray reduced values with a rolling window applied along the time dimension.
    """
    if window_length is not None:
        kwargs.update({str(time_dim): window_length})
    return _rolling_reduce(dataarray, **kwargs)



"""Temporary deprecated sub-module namespace."""

import warnings

from earthkit.transforms._aggregate import reduce, resample, rolling_reduce

from . import climatology, ensemble, general, spatial, temporal

warnings.warn(
    "The 'earthkit.transforms.aggregate' module is deprecated and will be removed "
    "in version 2.X of earthkit.transforms. Please import the from earthkit.transforms, e.g.: "
    "from earthkit.transforms import spatial",
    FutureWarning,
    stacklevel=2,
)

__all__ = [
    "climatology",
    "ensemble",
    "general",
    "spatial",
    "temporal",
    "reduce",
    "resample",
    "rolling_reduce",
]



"""Ensemble transformations for earthkit data objects.

Typically this is done with an xarray representation of data.
"""

from earthkit.transforms.ensemble._aggregate import max, mean, min, reduce, std, sum

# Backwards compatibility, necessary?
standard_deviation = std

__all__ = ["mean", "std", "min", "max", "sum", "reduce", "standard_deviation"]



"""Temporal transformations for earthkit data objects.

Typically this is done with an xarray representation of data. Some pandas methods are used
for indexing and selecting data.
"""

from earthkit.transforms.temporal._aggregate import (
    daily_max,
    daily_mean,
    daily_median,
    daily_min,
    daily_reduce,
    daily_std,
    daily_sum,
    max,
    mean,
    median,
    min,
    monthly_max,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_reduce,
    monthly_std,
    monthly_sum,
    reduce,
    rolling_reduce,
    standardise_time,
    std,
    sum,
)

__all__ = [
    "max",
    "mean",
    "median",
    "min",
    "reduce",
    "rolling_reduce",
    "std",
    "sum",
    "daily_max",
    "daily_mean",
    "daily_median",
    "daily_min",
    "daily_reduce",
    "daily_std",
    "daily_sum",
    "monthly_max",
    "monthly_mean",
    "monthly_median",
    "monthly_min",
    "monthly_reduce",
    "monthly_std",
    "monthly_sum",
    "standardise_time",
]



import numpy as np
import pytest
import xarray as xr
from earthkit.transforms import reduce, rolling_reduce
from earthkit.transforms._aggregate import (
    _reduce_dataarray,
    _rolling_reduce_dataarray,
)
from earthkit.transforms._tools import nanaverage


# Helper function to create a dummy DataArray
def create_dataarray(weights=None):
    data = xr.DataArray([[1, np.nan], [1, 2]], dims=("x", "y"))
    if weights is not None:
        data = data.weighted(weights=weights)
    return data


@pytest.mark.parametrize(
    "how, expected_result_method",
    (
        ["mean", np.nanmean],
        ["numpy.nanmean", np.nanmean],
        ["nanaverage", nanaverage],
        ["average", nanaverage],
        ["stddev", np.nanstd],
        ["std", np.nanstd],
        ["stdev", np.nanstd],
        ["sum", np.nansum],
        ["max", np.nanmax],
        ["min", np.nanmin],
        ["median", np.nanmedian],
    ),
)
# Test case for reducing without weights using built-in method
def test_reduce_dataarray(how, expected_result_method):
    dataarray = create_dataarray()
    result = _reduce_dataarray(dataarray, how=how)
    assert isinstance(result, xr.DataArray)
    assert result.shape == ()  # Ensure reduced to scalar
    assert result.values == expected_result_method(dataarray.values)

    result = reduce(dataarray, how)
    assert isinstance(result, xr.DataArray)
    assert result.shape == ()  # Ensure reduced to scalar
    assert result.values == expected_result_method(dataarray.values)

    result = reduce(xr.Dataset({"test": dataarray}), how)
    assert isinstance(result, xr.Dataset)
    assert result["test"].shape == ()  # Ensure reduced to scalar
    assert result["test"].values == expected_result_method(dataarray.values)

    result = reduce(xr.Dataset({"test": dataarray}), how, how_label=how)
    assert isinstance(result, xr.Dataset)
    assert result[f"test_{how}"].shape == ()  # Ensure reduced to scalar
    assert result[f"test_{how}"].values == expected_result_method(dataarray.values)


@pytest.mark.parametrize(
    "how",
    (
        "numpy.mean",
        "numpy.average",
    ),
)
# Test case for reducing with methods that should not ignore nans
def test_reduce_dataarray_nan(how):
    dataarray = create_dataarray()
    result = _reduce_dataarray(dataarray, how=how)
    assert isinstance(result, xr.DataArray)
    assert result.shape == ()  # Ensure reduced to scalar
    assert np.isnan(result.values)


@pytest.mark.parametrize(
    "how, expected_result_method",
    (
        ["mean", lambda x: x.mean()],
        ["nanmean", lambda x: x.mean()],
        ["average", lambda x: x.mean()],
        ["stddev", lambda x: x.std()],
        ["std", lambda x: x.std()],
        ["stdev", lambda x: x.std()],
    ),
)
# Test case for reducing with weights
def test_reduce_dataarray_with_weights(how, expected_result_method):
    dataarray = create_dataarray()
    weights = xr.DataArray([1, 2], dims="x", name="x")  # Dummy weights
    weighted_dataarray = create_dataarray(weights=weights)

    result = _reduce_dataarray(dataarray, how=how, weights=weights)
    assert isinstance(result, xr.DataArray)
    assert result.shape == ()  # Ensure reduced to scalar
    assert result.values == expected_result_method(weighted_dataarray)

    result = reduce(dataarray, how, weights=weights)
    assert isinstance(result, xr.DataArray)
    assert result.shape == ()  # Ensure reduced to scalar
    assert result.values == expected_result_method(weighted_dataarray)

    result = reduce(xr.Dataset({"test": dataarray}), how, weights=weights)
    assert isinstance(result, xr.Dataset)
    assert result["test"].shape == ()  # Ensure reduced to scalar
    assert result["test"].values == expected_result_method(weighted_dataarray)

    result = reduce(xr.Dataset({"test": dataarray}), how, weights=weights, how_label=how)
    assert isinstance(result, xr.Dataset)
    assert result[f"test_{how}"].shape == ()  # Ensure reduced to scalar
    assert result[f"test_{how}"].values == expected_result_method(weighted_dataarray)


# Helper function to create a dummy DataArray
def create_dataarray_rolling(weights=None):
    data = xr.DataArray(np.arange(25).reshape([5, 5]), dims=("x", "y"), name="test")
    if weights is not None:
        data = data.weighted(weights=weights)
    return data


@pytest.mark.parametrize(
    "how, expected_result",
    (
        ["mean", 2.5],
        ["numpy.nanmean", 2.5],
        ["nanaverage", 2.5],
        ["average", 2.5],
        ["stddev", 2.5],
        ["std", 2.5],
        ["stdev", 2.5],
        ["sum", 5],
        ["max", 5],
        ["min", 0],
        ["median", 2.5],
    ),
)
# Test case for reducing without weights using built-in method
def test_rolling_reduce_dataarray(how, expected_result):
    dataarray = create_dataarray_rolling()
    result = _rolling_reduce_dataarray(
        dataarray, how_reduce=how, x=2, center=True, chunk=False, how_dropna="any"
    )
    assert isinstance(result, xr.DataArray)
    assert result.shape == (4, 5)
    assert result.values[0, 0] == expected_result

    result = rolling_reduce(dataarray, how_reduce=how, x=2, center=True, chunk=False, how_dropna="any")
    assert isinstance(result, xr.DataArray)
    assert result.shape == (4, 5)
    assert result.values[0, 0] == expected_result

    result = rolling_reduce(
        xr.Dataset({"test": dataarray}), how_reduce=how, x=2, center=True, chunk=False, how_dropna="any"
    )
    assert isinstance(result, xr.Dataset)
    assert result["test"].shape == (4, 5)
    assert result["test"].values[0, 0] == expected_result

    result = rolling_reduce(
        xr.Dataset({"test": dataarray}),
        how_reduce=how,
        x=2,
        center=True,
        chunk=False,
        how_dropna="any",
        how_label=how,
    )
    assert isinstance(result, xr.Dataset)
    assert result[f"test_{how}"].shape == (4, 5)
    assert result[f"test_{how}"].values[0, 0] == expected_result



"""Spatial transformations for earthkit data objects.

Typically this is done with an xarray representation of data and a geopandas representation of geometries.
"""

from earthkit.transforms.spatial._aggregate import mask, reduce

__all__ = ["mask", "reduce"]



"""Module for preservation of the legacy API, it will be removed in version 2.X of earthkit.transforms."""
from earthkit.transforms.temporal._aggregate import (
    daily_max,
    daily_mean,
    daily_median,
    daily_min,
    daily_reduce,
    daily_std,
    daily_sum,
    max,
    mean,
    median,
    min,
    monthly_max,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_reduce,
    monthly_std,
    monthly_sum,
    reduce,
    rolling_reduce,
    standardise_time,
    std,
    sum,
)

__all__ = [
    standardise_time,
    reduce,
    mean,
    median,
    min,
    max,
    std,
    sum,
    daily_reduce,
    daily_mean,
    daily_median,
    daily_max,
    daily_min,
    daily_std,
    daily_sum,
    monthly_reduce,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_max,
    monthly_std,
    monthly_sum,
    rolling_reduce,
]
